def evaluator(cls):
    """
    Decorator to mark a class as an evaluator within the `afnio` framework.

    This decorator is intended to be applied to classes inheriting from `Function`.
    It sets an internal attribute `_is_evaluator` on the class to `True`, allowing
    the autodiff engine to recognize the class as an evaluator. This designation
    enables the evaluator’s `backward()` function to be called without any input
    arguments when computing gradients.

    Evaluators are responsible for assessing predictions using either deterministic
    or language model-based approaches. By using this decorator, users can build
    their own evaluator classes without worrying about internal setup details.

    Args:
        cls (type): The class being decorated, which should inherit from `Function`.

    Returns:
        type: The same class with the `_is_evaluator` attribute set to `True`.

    Example:
        >>> @evaluator
        >>> class MyCustomEvaluator(Function):
        >>>     @staticmethod
        >>>     def forward(ctx, prediction, target):
        >>>         # Evaluation logic
        >>>         pass
        >>>     @staticmethod
        >>>     def backward(ctx, grad_output):
        >>>         # Backpropagation logic
        >>>         pass
    """
    cls._is_evaluator = True
    return cls



from typing import Dict, List, Optional, Tuple, Union

from afnio._utils import (
    MultiTurnMessages,
)
from afnio._variable import Variable
from afnio.autodiff.function import Function
from afnio.models import ChatCompletionModel


class ChatCompletion(Function):
    r"""
    Implements a chat completion operation using the specified lanuage model within
    the `afnio` framework, supporting automatic differentiation.

    This class inherits from `Function` and requires both the `forward` and
    `backward` methods to be defined.

    ### Features:
    - **Mini-Batching**: Processes multiple input dictionaries simultaneously to improve
        throughput.
    - **Asynchronous Execution**: Both the forward and backward passes are optimized to
        run asynchronous calls for each mini-batch, reducing latency.
    - **Gradient Computation**: Supports automatic differentiation for all `Variables`
        in `messages` and `inputs` arguments, maintaining the order of gradients.

    The `ChatCompletion` function generates a `Variable` responses by passing a
    composite prompt, built from a list of `messages` and optional `inputs`, to the
    `forward_model_client`. Each message is a dictionary with a 'role' (e.g., 'system',
    'user') and a list of `Variable` objects as 'content'. `inputs` is a dictionary
    containing strings, list of strings or `Variable`s providing dynamic values to fill
    placeholders within message templates. If `inputs` contain lists of strings or
    `Variable`s which `.data` field is a list, the response's `.data` field will be a
    list, corresponding to the batched results. Otherwise, the `.data` field will be a
    scalar string. Additional behavior, such as temperature or token limits, can be
    customized through `completion_args`.

    Example with scalar inputs:
        >>> system = Variable(
        ...     "You are a helpful assistant.",
        ...     role="system instruction",
        ...     requires_grad=True
        ... )
        >>> user = Variable("Translate 'Hello' to {language}.", role="user query")
        >>> messages = [
        ...     {"role": "system", "content": [system]},
        ...     {"role": "user", "content": [user]},
        ... ]
        >>> inputs = {"language": Variable("Italian", role="language")}
        >>> response = ChatCompletion.apply(
        ...     model_client,
        ...     messages,
        ...     inputs=inputs,
        ...     temperature=0.7
        ... )
        >>> print(response.data)
        'Ciao'
        'Hola'
        >>> feedback = Variable("Use only capital letters.", role="feedback")
        >>> response.backward(feedback)
        >>> system.grad[0].data
        'The system instruction should enforce the use of capital letters only.'

    Example with batched inputs:
        >>> system = Variable(
        ...     "You are a helpful assistant.",
        ...     role="system instruction",
        ...     requires_grad=True
        ... )
        >>> user = Variable("Translate 'Hello' to {language}.", role="user query")
        >>> messages = [
        ...     {"role": "system", "content": [system]},
        ...     {"role": "user", "content": [user]},
        ... ]
        >>> inputs = {
        ...     "language": [
        ...         Variable("Italian", role="language"),
        ...         Variable("Spanish", role="language")
        ...     ]
        ... }
        >>> response = ChatCompletion.apply(
        ...     model_client,
        ...     messages,
        ...     inputs=inputs,
        ...     temperature=0.7
        ... )
        >>> print(response.data)
        ['Ciao', 'Hola']
    """

    @staticmethod
    def forward(
        ctx,
        forward_model_client: Optional[ChatCompletionModel],
        messages: MultiTurnMessages,
        inputs: Optional[Dict[str, Union[str, List[str], Variable]]] = None,
        **completion_args,
    ) -> Variable:
        raise NotImplementedError(
            "ChatCompletion.forward is implemented on the server. "
            "Client-side execution is not supported."
        )

    @staticmethod
    def backward(ctx, grad_output: Variable) -> Tuple[Optional[Variable], ...]:
        raise NotImplementedError(
            "ChatCompletion.backward is implemented on the server. "
            "Client-side execution is not supported."
        )



from typing import TYPE_CHECKING, Dict, Optional

if TYPE_CHECKING:
    from afnio.autodiff.graph import GradientEdge, Node

# Registry mapping node_id to Node instance
NODE_REGISTRY: Dict[str, "Node"] = {}


def register_node(node: "Node"):
    """
    Register a Node instance in the local registry.

    Args:
        node (Node): The Node instance to register.
    """
    if node.node_id:
        NODE_REGISTRY[node.node_id] = node


def get_node(node_id: str) -> Optional["Node"]:
    """
    Retrieve a Node instance from the registry by its node_id.

    Args:
        node_id (str): The unique identifier of the Node.

    Returns:
        Node or None: The Node instance if found, else None.
    """
    return NODE_REGISTRY.get(node_id)


def create_node(data: dict) -> "Node":
    """
    Create and register a Node from serialized data received from the server.

    Args:
        data (dict): Serialized node data with keys 'name' and 'node_id'.

    Returns:
        Node: The created and registered Node instance.
    """
    from afnio._variable import _allow_grad_fn_assignment
    from afnio.autodiff.graph import Node
    from afnio.tellurio._variable_registry import (
        PENDING_GRAD_FN_ASSIGNMENTS,
        suppress_variable_notifications,
    )

    node = Node()
    node._name = data["name"]
    node.node_id = data["node_id"]
    register_node(node)

    # After registering, resolve any pending grad_fn assignments
    pending_vars = PENDING_GRAD_FN_ASSIGNMENTS.pop(node.node_id, [])
    for var in pending_vars:
        with suppress_variable_notifications(), _allow_grad_fn_assignment():
            var._pending_grad_fn_id = None
            var.grad_fn = node

    return node


def create_and_append_edge(data: dict) -> "GradientEdge":
    """
    Create a GradientEdge from serialized data
    and append it to from_node.next_functions.

    Note:
        The edge is appended to from_node.next_functions and points to to_node.
        This follows the backward pass convention.

    Args:
        data (dict): Serialized edge data with keys 'from_node_id', 'to_node_id',
          and 'output_nr'.

    Returns:
        GradientEdge: The created GradientEdge instance.
    """
    from afnio.autodiff.graph import GradientEdge

    from_node_id = data["from_node_id"]
    to_node_id = data["to_node_id"]
    from_node = get_node(from_node_id)
    to_node = get_node(to_node_id)
    output_nr = data["output_nr"]

    if not from_node:
        raise ValueError(f"from_node with id '{from_node_id}' not found in registry.")
    if to_node_id is not None and to_node is None:
        raise ValueError(f"to_node with id '{to_node_id}' not found in registry.")
    # If to_node_id=None, to_node=None (leaf Variable that doesn't require gradient)

    edge = GradientEdge(node=to_node, output_nr=output_nr)
    from_node.next_functions = from_node.next_functions + (edge,)
    return edge



import inspect
import pickle
import types
import uuid
from typing import Any, Callable, Dict, List, Optional

from afnio._variable import Variable
from afnio.models import ChatCompletionModel
from afnio.models.model import BaseModel
from afnio.tellurio._callable_registry import register_callable
from afnio.tellurio._model_registry import get_model
from afnio.tellurio._variable_registry import get_variable

MultiTurnMessages = List[Dict[str, List[Variable]]]


def _validate_typed_sequence(
    input_value, item_type, allow_tuple=False, allow_single=False, input_name=None
):
    """
    Validates that the input is a list (or optionally a tuple or single instance) of a
    specified type.

    Args:
        input_value: The value to validate.
        item_type: The expected type for each element.
        allow_tuple: If True, allows a tuple instead of a list.
        allow_single: If True, allows a single instance instead of a list.
        input_name: The name of the variable being validated (used for error messages).
            If `_validate_typed_sequence()` is not nested within other functions this
            argument is not necessary as it gets automatically derived. Pass this
            argument when `_validate_typed_sequence()` is nested one or multiple times.

    Raises:
        TypeError: If the input is not a list of the specified type (or a single
            instance if allowed).
    """
    if input_name is None:
        caller_locals = inspect.currentframe().f_back.f_locals
        variable_name = [
            name for name, value in caller_locals.items() if value is input_value
        ]
        input_name = variable_name[0] if variable_name else "input"

    if isinstance(input_value, item_type) and allow_single:
        return

    valid_types = (list, tuple) if allow_tuple else (list,)
    type_names = ", ".join(t.__name__ for t in valid_types)
    if not isinstance(input_value, valid_types):
        raise TypeError(
            f"`{input_name}` expects a ({type_names}) of type {item_type.__name__}, "
            f"but got {type(input_value).__name__}."
        )
    if not all(isinstance(item, item_type) for item in input_value):
        raise TypeError(
            f"All elements in `{input_name}` must be of type {item_type.__name__}, "
            f"but got {[type(item).__name__ for item in input_value]}."
        )


def _validate_multi_turn_messages(messages: MultiTurnMessages):
    """
    Validates the structure and content of the `messages` parameter for
    multi-turn composition.

    Args:
        messages (MultiTurnMessages): A list of dictionaries, each representing a chat
            turn with a 'role' and a list of `Variable` objects in the 'content'.

    Raises:
        ValueError: If any message lacks 'role' or 'content' keys.
        TypeError: If 'role' in any message is not a string.
        TypeError: If 'content' in any message is not a list of `Variable` objects.
    """
    _validate_typed_sequence(messages, dict)

    for msg in messages:
        if "role" not in msg or "content" not in msg:
            raise ValueError(
                f"Each message dictionary must contain 'role' and 'content' keys, "
                f"but got {msg.keys()}."
            )
        if not isinstance(msg["role"], str):
            raise TypeError(
                f"'role' must be a string, but got {type(msg['role']).__name__}."
            )
        _validate_typed_sequence(msg["content"], Variable, input_name='msg["content"]')


def is_multi_turn_messages(input_value: MultiTurnMessages):
    """
    Determines whether the given input conforms to the structure of multi-turn messages.

    This function validates if the `input_value` adheres to the expected format for
    multi-turn chat messages, which should be a list of dictionaries where each
    dictionary contains a 'role' (a string) and a 'content' (a list of `Variable`
    objects).

    Args:
        input_value (MultiTurnMessages): The input value to validate.

    Returns:
        bool: Returns `True` if the input is valid multi-turn messages,
            otherwise `False`.
    """
    try:
        _validate_multi_turn_messages(input_value)
        return True
    except TypeError:
        return False


def _validate_function(func: Callable[..., Any]) -> None:
    """
    Validate whether `func` is a valid standalone function that can be safely pickled.

    Allowed functions:
    - Standalone `def` functions (global or module-level)
    - Built-in functions
    - `staticmethod`s (if extracted correctly)

    The function must:
    - Not be a lambda
    - Not be a method (bound or unbound)
    - Not be a callable object (e.g., a class instance with `__call__`)
    - Not be a nested function or closure
    - Be picklable

    Args:
        func (Callable[..., Any]): The function to validate.

    Raises:
        TypeError: If `func` is not a valid function.
        ValueError: If `func` cannot be safely pickled.
    """

    # Reject methods (bound or unbound)
    if isinstance(func, types.MethodType):
        raise TypeError(
            f"Invalid function: '{func.__name__}' is a method (bound or unbound). "
            f"Methods are tied to a class or instance, which prevents pickling. "
            f"If this function does not require 'self' or 'cls', "
            f"consider making it a @staticmethod."
        )

    # Ensure it's a function (user-defined or built-in)
    if not isinstance(func, (types.FunctionType, types.BuiltinFunctionType)):
        raise TypeError(
            f"Invalid function: '{func}'. Expected a standalone function, "
            f"but got a {type(func).__name__}."
        )

    # Reject lambdas
    if func.__name__ == "<lambda>":
        raise TypeError(
            "Invalid function: Lambda functions cannot be pickled and are not allowed."
        )

    # Reject nested functions and closures (but only check __closure__ if it exists)
    if hasattr(func, "__closure__") and func.__closure__:
        raise TypeError(
            f"Invalid function: '{func.__name__}' is a closure or a nested function. "
            f"Closures capture variables from their enclosing scope, making them "
            f"unpicklable. Consider rewriting it as a standalone function "
            f"or a callable class."
        )

    # Test picklability
    try:
        pickle.dumps(func)
    except (pickle.PickleError, AttributeError) as e:
        raise ValueError(
            f"Invalid function: '{func.__name__}' cannot be pickled. Error: {e}"
        )


def _is_valid_function(func: Optional[Callable[..., Any]]) -> bool:
    """
    Check if `func` is a valid standalone function that can be safely pickled.

    Allowed functions:
    - Standalone `def` functions (global or module-level)
    - Built-in functions
    - `staticmethod`s (if extracted correctly)

    The function must:
    - Not be a lambda
    - Not be a method (bound or unbound)
    - Not be a callable object (e.g., a class instance with `__call__`)
    - Not be a nested function or closure
    - Be picklable

    Args:
        func (Callable[..., Any] or None): The function to validate.
            `None` is considered valid.

    Returns:
        bool: True if the function is valid, False otherwise.
    """
    if func is None:
        return True  # `None` is considered valid

    try:
        _validate_function(func)
        return True
    except (TypeError, ValueError):
        return False


def _serialize_arg(arg: Any) -> Any:
    """
    Recursively serialize an argument for RPC transmission.

    Handles:
    - Parameter: serializes as a dict with a type tag and variable_id.
    - Variable: serializes as a dict with a type tag and variable_id.
    - ChatCompletionModel: serializes as a dict with a type tag and model_id.
    - Callable: registers the callable and serializes as a dict with a type tag
      and callable_id.
    - list/tuple: recursively serializes each element.
    - dict: recursively serializes each value.
    - Primitives (str, int, float, bool, None): returned as-is.

    Callables are not currently supported and will raise if encountered.
    """
    from afnio.cognitive.parameter import Parameter

    if isinstance(arg, Parameter):
        return {
            "__parameter__": True,
            "variable_id": arg.variable_id,
        }
    elif isinstance(arg, Variable):
        return {
            "__variable__": True,
            "variable_id": arg.variable_id,
        }
    elif isinstance(arg, ChatCompletionModel):
        return {
            "__model_client__": True,
            "model_id": arg.model_id,
        }
    elif callable(arg):
        # Register the callable and generate a unique ID
        callable_id = str(uuid.uuid4())
        register_callable(callable_id, arg)
        return {
            "__callable__": True,
            "callable_id": callable_id,
        }
    elif isinstance(arg, list):
        return [_serialize_arg(a) for a in arg]
    elif isinstance(arg, tuple):
        return tuple(_serialize_arg(a) for a in arg)
    elif isinstance(arg, dict):
        return {k: _serialize_arg(v) for k, v in arg.items()}
    elif isinstance(arg, (str, int, float, bool)) or arg is None:
        return arg
    else:
        raise TypeError(
            f"Cannot serialize object of type {type(arg).__name__}: {arg!r}"
        )


def _deserialize_output(obj: Any) -> Any:
    """
    Recursively deserialize objects received from the client.

    This function converts serialized representations of Variables, Parameters and model
    clients (as sent by the server) back into their corresponding client-side objects
    using the session registries. It handles lists, tuples, and dictionaries
    recursively, and leaves primitive types unchanged.

    Args:
        obj: The serialized argument to deserialize. Can be a dict, list, tuple,
            or primitive.

    Returns:
        The deserialized object, with Variables, Parameters amd LM models resolved from
        the session registries.

    Raises:
        ValueError: If a referenced variable or model cannot be found in the context.
        TypeError: If the input type is not supported.
    """
    from afnio.cognitive.parameter import Parameter

    if isinstance(obj, dict):
        if obj.get("__variable__") and "variable_id" in obj:
            variable_id = obj["variable_id"]
            try:
                variable = get_variable(variable_id)
                if not variable and not isinstance(variable, Variable):
                    raise ValueError(
                        f"Variable with variable_id {variable_id!r} not found"
                    )
            except KeyError:
                raise ValueError(f"Unknown variable_id: {variable_id!r}")
            return variable
        elif obj.get("__parameter__") and "variable_id" in obj:
            variable_id = obj["variable_id"]
            try:
                parameter = get_variable(variable_id)
                if not parameter and not isinstance(parameter, Parameter):
                    raise ValueError(
                        f"Parameter with variable_id {variable_id!r} not found"
                    )
            except KeyError:
                raise ValueError(f"Unknown variable_id: {variable_id!r}")
            return parameter
        elif obj.get("__model_client__") and "model_id" in obj:
            model_id = obj["model_id"]
            try:
                model = get_model(model_id)
                if not model and not isinstance(model, BaseModel):
                    raise ValueError(f"Model with model_id {model_id!r} not found")
            except KeyError:
                raise ValueError(f"Unknown model_id: {model_id!r}")
            return model
        else:
            return {k: _deserialize_output(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [_deserialize_output(v) for v in obj]
    elif isinstance(obj, tuple):
        return tuple(_deserialize_output(v) for v in obj)
    elif isinstance(obj, (str, int, float, bool, type(None))):
        return obj
    else:
        raise TypeError(f"Cannot deserialize object of type {type(obj)}")



import builtins
import json
import tempfile

import pytest

from afnio.tellurio.consent import check_consent
from afnio.tellurio.utils import get_config_path


class TestCheckConsent:
    """Test suite for check_consent function."""

    def test_env_true(self, monkeypatch):
        """
        Test that check_consent passes when ALLOW_API_KEY_SHARING is 'true' in env.
        """
        monkeypatch.setenv("ALLOW_API_KEY_SHARING", "true")
        # Should not raise
        check_consent()

    def test_env_invalid(self, monkeypatch):
        """
        Test that check_consent raises ValueError for invalid env value.
        """
        monkeypatch.setenv("ALLOW_API_KEY_SHARING", "maybe")
        with pytest.raises(ValueError):
            check_consent()

    def test_config_true(self, monkeypatch):
        """
        Test that check_consent passes when config file allows sharing.
        """
        config_path = get_config_path()
        with open(config_path, "w") as f:
            json.dump({"allow_api_key_sharing": True}, f)
        monkeypatch.delenv("ALLOW_API_KEY_SHARING", raising=False)
        # Should not raise
        check_consent()

    def test_prompt_yes(self, monkeypatch):
        """
        Test that check_consent passes when user inputs 'yes' interactively.
        """
        config_path = get_config_path()
        with open(config_path, "w") as f:
            json.dump({"allow_api_key_sharing": "not-true"}, f)
        monkeypatch.setattr("afnio.tellurio.consent._is_interactive", lambda: True)
        monkeypatch.delenv("ALLOW_API_KEY_SHARING", raising=False)
        monkeypatch.setattr(builtins, "input", lambda _: "yes")
        # Should not raise
        check_consent()

    def test_prompt_always(self, monkeypatch):
        """
        Test that check_consent passes and updates config when user inputs 'always'.
        """
        config_path = get_config_path()
        with open(config_path, "w") as f:
            json.dump({"allow_api_key_sharing": "not-true"}, f)
        monkeypatch.setattr("afnio.tellurio.consent._is_interactive", lambda: True)
        monkeypatch.delenv("ALLOW_API_KEY_SHARING", raising=False)
        monkeypatch.setattr(builtins, "input", lambda _: "always")
        check_consent()
        with open(config_path) as f:
            config = json.load(f)
        assert config["allow_api_key_sharing"] is True

    def test_prompt_no(self, monkeypatch):
        """
        Test that check_consent raises when user inputs 'no' interactively.
        """
        # Simulate interactive environment
        config_path = get_config_path()
        with open(config_path, "w") as f:
            json.dump({"allow_api_key_sharing": "not-true"}, f)
        monkeypatch.setattr("afnio.tellurio.consent._is_interactive", lambda: True)
        monkeypatch.delenv("ALLOW_API_KEY_SHARING", raising=False)
        monkeypatch.setattr(builtins, "input", lambda _: "no")
        with pytest.raises(
            RuntimeError,
            match="Consent to share your LM API key was not given. Aborting operation.",
        ):
            check_consent()

    def test_noninteractive(self, monkeypatch):
        """
        Test that check_consent raises in non-interactive environment with no consent.
        """
        # Simulate non-interactive environment
        monkeypatch.setattr("afnio.tellurio.consent._is_interactive", lambda: False)
        monkeypatch.delenv("ALLOW_API_KEY_SHARING", raising=False)
        monkeypatch.setattr(
            "afnio.tellurio.utils.get_config_path", lambda: tempfile.mktemp()
        )
        with pytest.raises(
            RuntimeError,
            match=(
                "Consent to share your API key with the server is required, "
                "but no consent was found and the environment is non-interactive. "
                "Set ALLOW_API_KEY_SHARING=True or update ~/.tellurio.config to allow."
            ),
        ):
            check_consent()



import asyncio
import threading


class EventLoopThread:
    """
    Runs an asyncio event loop in a dedicated background thread.

    This class enables synchronous code to execute asynchronous coroutines
    by submitting them to a persistent event loop running outside the main thread.
    It manages the lifecycle of the event loop and its thread, providing methods
    to run coroutines and to shut down the loop cleanly.
    """

    def __init__(self):
        # Create a new asyncio event loop (not the default one)
        self.loop = asyncio.new_event_loop()
        # Start the event loop in a dedicated daemon thread
        self.thread = threading.Thread(target=self.loop.run_forever, daemon=True)
        self.thread.start()

    def run(self, coro):
        """
        Submit a coroutine to the background event loop and block until it completes.

        Args:
            coro: The coroutine to run.

        Returns:
            The result of the coroutine.
        """
        # Schedule the coroutine to run in the background event loop
        future = asyncio.run_coroutine_threadsafe(coro, self.loop)
        # Block until the coroutine completes and return its result
        return future.result()

    def shutdown(self):
        """
        Stop the event loop and wait for the background thread to finish.
        """
        self.loop.call_soon_threadsafe(self.loop.stop)
        self.thread.join()


# Singleton instance of the background event loop thread
_event_loop_thread = EventLoopThread()


def run_in_background_loop(coro):
    """
    Submit a coroutine to the background event loop and return its result.
    """
    return _event_loop_thread.run(coro)



import logging
from typing import Callable, Dict, List, Optional, Tuple, Union

from afnio._utils import (
    MultiTurnMessages,
    _deserialize_output,
    _serialize_arg,
)
from afnio._variable import Variable
from afnio.logging_config import configure_logging
from afnio.models import ChatCompletionModel
from afnio.tellurio._eventloop import run_in_background_loop
from afnio.tellurio._variable_registry import (
    suppress_variable_notifications,
)
from afnio.tellurio.client import get_default_client

from .optimizer import Optimizer, ParamsT, _extract_variable_ids, _wait_for_variable

# Configure logging
configure_logging()
logger = logging.getLogger(__name__)


# Suppress notifications for variable changes during modules initialization
# as the WebSocket connection is not established yet
with suppress_variable_notifications():
    # TGD_MESSAGES is only a placeholder that will be replaced on the server side by the
    # actual system prompt and user instruction
    TGD_MESSAGES = [
        {
            "role": "system",
            "content": [
                Variable(
                    data="Placeholder for Textual Gradient Descent optimizer system prompt",  # noqa: E501
                    role="Textual Gradient Descent optimizer system prompt",
                )
            ],
        },
        {
            "role": "user",
            "content": [
                Variable(
                    data="Placeholder for Textual Gradient Descent optimizer user prompt",  # noqa: E501
                    role="Textual Gradient Descent optimizer user prompt",
                )
            ],
        },
    ]


class TGD(Optimizer):
    def __init__(
        self,
        params: ParamsT,
        model_client: Optional[ChatCompletionModel],
        messages: MultiTurnMessages = TGD_MESSAGES,
        inputs: Optional[Dict[str, Union[str, Variable]]] = None,
        constraints: Optional[List[Union[str, Variable]]] = None,
        momentum: int = 0,
        **completion_args,
    ):
        """
        Textual Gradient Descent (TGD) optimizer.

        Args:
            params (ParamsT): Iterable of parameters to optimize or dicts defining
                parameter groups.
            model_client (Optional[ChatCompletionModel]): LM model client used
                for optimization.
            messages (MultiTurnMessages): Messages for multi-turn interactions. It
                typically defines the optimizer system prompt and user instruction.
                In-context examples (shots) can be added as well.
            inputs (Optional[Dict[str, Union[str, Variable]]]): Dynamic values to fill
                placeholders within message templates
            constraints (Optional[List[Union[str, Variable]]]): A list of natural
                language constraints for optimization.
            momentum (int, optional): Momentum window size. Tracks the last `momentum`
                gradients, which helps accelerate updates in the right direction and
                dampen oscillations. Defaults to 0.
            completion_args (Dict[str, Any], optional): Additional arguments to pass to
                the model client when generating text completions. Defaults to an
                empty dictionary.
        """
        # Workaround to trigger TGD_MESSAGES registration with the server
        # and store related variable_ids on the client side
        if messages is TGD_MESSAGES:
            messages = [
                {
                    "role": "system",
                    "content": [
                        Variable(
                            data="Placeholder for Textual Gradient Descent optimizer system prompt",  # noqa: E501
                            role="Textual Gradient Descent optimizer system prompt",
                        )
                    ],
                },
                {
                    "role": "user",
                    "content": [
                        Variable(
                            data="Placeholder for Textual Gradient Descent optimizer user prompt",  # noqa: E501
                            role="Textual Gradient Descent optimizer user prompt",
                        )
                    ],
                },
            ]

        defaults = dict(
            model_client=model_client,
            messages=messages,
            inputs=inputs or {},
            constraints=constraints or [],
            momentum=momentum,
            completion_args=completion_args,
        )
        super().__init__(params, defaults)

    def step(
        self, closure: Optional[Callable] = None
    ) -> Optional[Tuple[Variable, Variable]]:
        """Performs a single optimization step.

        Args:
            closure (Optional[Callable]): A closure that reevaluates the model
                and returns the loss.

        Returns:
            Optional[Tuple[Variable, Variable]]: The loss if `closure` is provided,
                otherwise None. The loss should return a numerical or textual score and
                a textual explanation, both wrapped as `Variable` objects
        """
        loss = closure() if closure else (None, None)
        super().step()
        return loss

    def _extract_variable_ids_from_state(self, state):
        """
        Extract only the variable_ids of deepcopied parameters (i.e., those generated
        on the server) from the optimizer state.

        Args:
            state (list): The serialized optimizer state as returned by the server.

        Returns:
            Set[str]: Set of variable_ids for deepcopied parameters.
        """
        var_ids = set()
        for entry in state:
            momentum_buffer = entry.get("value", {}).get("momentum_buffer", [])
            for buf_entry in momentum_buffer:
                if (
                    isinstance(buf_entry, list)
                    and len(buf_entry) > 0
                    and isinstance(buf_entry[0], dict)
                    and "variable_id" in buf_entry[0]
                ):
                    var_ids.add(buf_entry[0]["variable_id"])
        return var_ids


# TODO: Fix error when passing str constraints
def tgd(
    params: List[Variable],
    grads: List[List[Variable]],
    momentum_buffer_list: List[Optional[List[Variable]]],
    model_client: Optional[ChatCompletionModel],
    messages: MultiTurnMessages,
    inputs: Optional[Dict[str, Union[str, Variable]]],
    constraints: Optional[List[Union[str, Variable]]],
    momentum: int,
    **completion_args,
):
    r"""Functional API that performs TGD (Textual Gradient Descent) algorithm
    computation.

    See :class:`~afnio.optim.SGD` for details.
    """
    # Set `_pending_data` for all parameters that will be optimized
    for p in params:
        p._pending_data = True
        logger.debug(f"Marked variable {p.variable_id!r} as pending for data update.")

    try:
        _, ws_client = get_default_client()

        payload = {
            "params": _serialize_arg(params),
            "grads": _serialize_arg(grads),
            "momentum_buffer_list": _serialize_arg(momentum_buffer_list),
            "model_client": _serialize_arg(model_client),
            "messages": _serialize_arg(messages),
            "inputs": _serialize_arg(inputs),
            "constraints": _serialize_arg(constraints),
            "momentum": momentum,
            "completion_args": _serialize_arg(completion_args),
        }

        response = run_in_background_loop(ws_client.call("run_optimizer_tgd", payload))
        logger.debug(f"TGD optimization request sent: {payload!r}")

        result = response.get("result", {})
        result_message = result.get("message")
        result_momentum_buffer_list = result.get("momentum_buffer_list", [])

        # Extract all variable_ids from the result_momentum_buffer_list
        # and wait for them to be registered in VARIABLE_REGISTRY
        all_var_ids = _extract_variable_ids(result_momentum_buffer_list)
        for var_id in all_var_ids:
            _wait_for_variable(var_id)

        des_momentum_buffer_list = _deserialize_output(result_momentum_buffer_list)

        # Convert [param, grads] lists to (param, grads) tuples
        for i, buffer in enumerate(des_momentum_buffer_list):
            des_momentum_buffer_list[i] = [
                tuple(pair) if isinstance(pair, list) and len(pair) == 2 else pair
                for pair in buffer
            ]

        if result_message != "Functional TGD optimization step executed successfully.":
            logger.error(
                f"Server did not return any data for functional TGD optimization: "
                f"payload={payload!r}, response={response!r}"
            )
            raise RuntimeError(
                "Failed to run functional TGD optimization: server did not return data."
            )

        # Update the momentum_buffer_list with the deserialized buffers
        momentum_buffer_list.clear()
        momentum_buffer_list.extend(des_momentum_buffer_list)

        logger.debug("Functional TGD optimization executed successfully")
    except Exception as e:
        logger.error(f"Failed to run functional TGD optimization: {e!r}")
        raise



from typing import List, Optional, Tuple, Union

from afnio._variable import Variable
from afnio.autodiff.function import Function


class Add(Function):
    r"""Implements an addition operation for `Variable` instances within the
    `afnio` framework, supporting automatic differentiation.

    This class inherits from `autodiff.Function` and requires both the `forward`
    and `backward` methods to be defined.

    The `Add` function supports both scalar and list `.data` fields:
    - **Scalars**: Adds numerical values (`int`, `float`) or concatenates strings.
    - **Lists**: Performs element-wise addition of corresponding elements from the lists.
      Lists must be of the same length.

    It automatically handles type-based operations:
    - For numerical data (`int`, `float`), it performs arithmetic addition.
    - For strings, it concatenates the values.
    - Mixed types (e.g., string and number) are converted appropriately before performing
      the addition.

    This operation also tracks `Variable` dependencies, enabling automatic gradient
    computation through backpropagation.

    Example with scalar inputs:
        >>> x = Variable(data="abc", role="first input", requires_grad=True)
        >>> y = Variable(data="def", role="second input", requires_grad=False)
        >>> result = Add.apply(x, y)
        >>> result.data
        'abcdef'
        >>> result.role
        'first input and second input'
        >>> result.requires_grad
        True
        >>> g = Variable(data="MY_FEEDBACK", role="add gradient")
        >>> result.backward(g)
        >>> x.grad[0].data
        'Here is the combined feedback we got for this specific first input and other variables: MY_FEEDBACK'
        >>> x.grad[0].role
        'feedback to first input'

    Example with batched inputs:
        >>> x = Variable(data=[1, 2, 3], role="first input", requires_grad=True)
        >>> y = Variable(data=[4, 5, 6], role="second input", requires_grad=False)
        >>> result = Add.apply(x, y)
        >>> result.data
        [5, 7, 9]
        >>> result.role
        'first input and second input'
        >>> result.requires_grad
        True
    """  # noqa: E501

    @staticmethod
    def forward(ctx, x: Variable, y: Variable) -> Variable:
        raise NotImplementedError(
            "Add.forward is implemented on the server. "
            "Client-side execution is not supported."
        )

    @staticmethod
    def backward(
        ctx, grad_output: Variable
    ) -> Tuple[Optional[Variable], Optional[Variable]]:
        raise NotImplementedError(
            "Add.backward is implemented on the server. "
            "Client-side execution is not supported."
        )


# TODO: enable summarization of `.data`` fields in `forward()` method using a LM
#       (ideally a small and fast SLM)
class Sum(Function):
    r"""
    Implements a summation operation for a list of `Variable` instances within the
    `afnio` framework, supporting automatic differentiation.

    This class inherits from `Function` and requires both the `forward` and
    `backward` methods to be defined.

    The `Sum` function aggregates the `.data`, `.role`, and `.requires_grad` attributes
    of all input `Variable` instances into a single `Variable`. It supports both
    scalar and list `.data` fields:
    - **Scalars**: Computes the arithmetic sum for numerical data (`int`, `float`)
      or concatenates all string values, wrapping each in `<ITEM></ITEM>` tags.
    - **Lists**: Aggregates the corresponding elements of the lists. For numerical
      data, it sums the corresponding elements. For string data, it concatenates them,
      wrapping each element in `<ITEM></ITEM>` tags.

    During backpropagation, the function distributes the gradient to all input
    `Variable` instances that require gradients.

    Example with scalar inputs:
        >>> x = Variable(data="abc", role="first input", requires_grad=True)
        >>> y = Variable(data="def", role="second input", requires_grad=False)
        >>> result = Sum.apply([x, y])
        >>> result.data
        '<ITEM>abc</ITEM><ITEM>def</ITEM>'
        >>> result.role
        'first input and second input'
        >>> result.requires_grad
        True
        >>> g = Variable(data="MY_FEEDBACK", role="add gradient")
        >>> result.backward(g)
        >>> x.grad[0].data
        'Here is the combined feedback we got for this specific first input and other variables: MY_FEEDBACK'
        >>> x.grad[0].role
        'feedback to first input'

    Example with batched inputs:
        >>> x = Variable(data=[1, 2, 3.5], role="first input", requires_grad=True)
        >>> y = Variable(data=[4, 5, 6], role="second input", requires_grad=False)
        >>> result = Sum.apply([x, y])
        >>> result.data
        [5, 7, 9.5]
        >>> result.role
        'first input and second input'
        >>> result.requires_grad
        True
    """  # noqa: E501

    @staticmethod
    def forward(ctx, x: List[Variable]) -> Variable:
        raise NotImplementedError(
            "Sum.forward is implemented on the server. "
            "Client-side execution is not supported."
        )

    @staticmethod
    def backward(ctx, grad_output: Variable) -> Tuple[Optional[Variable], ...]:
        raise NotImplementedError(
            "Sum.backward is implemented on the server. "
            "Client-side execution is not supported."
        )


class Split(Function):
    r"""
    Implements a split operation for `Variable` instances within the
    `afnio` framework, supporting automatic differentiation.

    This class inherits from `Function` and requires both the `forward` and
    `backward` methods to be defined.

    The `Split` function divides the `.data` of the input `Variable` into multiple parts
    using a specified delimiter `sep`. If `maxsplit` is specified, the split operation
    is limited to a maximum number of splits. It handles both scalar and list `.data`
    fields:
    - **Scalars**: The scalar `.data` (a single string) is split into substrings
      based on the specified `sep` and `maxsplit` parameters.
    - **Lists**: Each element of the list `.data` (strings) is split individually. If
      splits of varying lengths occur, shorter splits are automatically padded with
      empty strings to ensure consistent dimensions.

    During backpropagation, feedback is collected and aggregated across all split parts.
    The combined feedback is propagated back to the original input `Variable`, allowing
    for the proper computation of gradients.

    Example with scalar inputs:
        >>> x = Variable(data="afnio is great!", role="sentence", requires_grad=True)
        >>> result = Split.apply(x, sep=" ", maxsplit=1)
        >>> [var.data for var in result]
        ['afnio', 'is great!']
        >>> result[0].role
        'split part 0 of sentence'
        >>> g_1 = Variable(data="MY_FIRST_FEEDBACK", role="gradient")
        >>> g_2 = Variable(data="MY_SECOND_FEEDBACK", role="gradient")
        >>> result[0].backward(g_1, retain_graph=True)
        >>> result[1].backward(g_2)
        >>> x.grad[0].data
        'Here is the combined feedback we got for this specific sentence and other variables: <ITEM>MY_FIRST_FEEDBACK</ITEM><ITEM></ITEM>'
        >>> x.grad[0].role
        'feedback to sentence'
        >>> x.grad[1].data
        'Here is the combined feedback we got for this specific sentence and other variables: <ITEM></ITEM><ITEM>MY_SECOND_FEEDBACK</ITEM>'
        >>> x.grad[1].role
        'feedback to sentence'

    Example with batched inputs:
        >>> x = Variable(
        ...     data=["afnio is great!", "Deep learning"],
        ...     role="sentences",
        ...     requires_grad=True
        ... )
        >>> result = Split.apply(x, sep=" ", maxsplit=2)
        >>> [var.data for var in result]
        [['afnio', 'Deep'], ['is', 'learning'], ['great!', '']]
        >>> g = Variable(data="MY_FEEDBACK", role="gradient")
        >>> result[1].backward(g)
        >>> x.grad[0].data
        'Here is the combined feedback we got for this specific sentences and other variables: <ITEM></ITEM><ITEM>MY_FEEDBACK</ITEM><ITEM></ITEM>'
        >>> x.grad[0].role
        'feedback to sentences'
    """  # noqa: E501

    @staticmethod
    def forward(
        ctx,
        x: Variable,
        sep: Optional[Union[str, Variable]] = None,
        maxsplit: Optional[Union[int, Variable]] = -1,
    ) -> Tuple[Variable]:
        raise NotImplementedError(
            "Split.forward is implemented on the server. "
            "Client-side execution is not supported."
        )

    # TODO: enable summarization of elements in `grad.data`` fields in `backward()`
    #       method using a LM (ideally a small and fast SLM)
    @staticmethod
    def backward(ctx, *grad_outputs: Variable) -> Tuple[Optional[Variable], None, None]:
        raise NotImplementedError(
            "Split.backward is implemented on the server. "
            "Client-side execution is not supported."
        )



import json
import logging
import os

import httpx
import keyring
from dotenv import load_dotenv

from afnio.logging_config import configure_logging
from afnio.tellurio.utils import get_config_path
from afnio.tellurio.websocket_client import TellurioWebSocketClient

# Configure logging
configure_logging()
logger = logging.getLogger(__name__)

# Load environment variables from .env
load_dotenv()

# Define the global default client instances
_default_client = None
_default_ws_client = None


def save_username(username):
    """
    Saves the username to a JSON configuration file.
    If the username is different from the one already stored, this function
    updates the 'username' field and clears all other preferences.
    Otherwise, it preserves all existing values.
    """
    config_path = get_config_path()
    # Load existing config if present, otherwise start with empty dict
    if os.path.exists(config_path):
        with open(config_path, "r") as f:
            try:
                config = json.load(f)
            except json.JSONDecodeError:
                config = {}
    else:
        config = {}

    # If username is different, clear all preferences except username
    if config.get("username") != username:
        config = {"username": username}
    else:
        config["username"] = username

    with open(config_path, "w") as f:
        json.dump(config, f, indent=2)


def load_username():
    """
    Loads the username from a JSON configuration file.
    This function reads the JSON file at the specified path and retrieves the username
    stored in it. If the file does not exist, it returns None.
    """
    config_path = get_config_path()
    if os.path.exists(config_path):
        with open(config_path, "r") as f:
            try:
                return json.load(f).get("username")
            except json.JSONDecodeError:
                logger.debug("Failed to decode JSON from the configuration file.")
                return None
    return None


class InvalidAPIKeyError(Exception):
    """Exception raised when the API key is invalid."""

    pass


class TellurioClient:
    """
    A client for interacting with the Tellurio backend.

    This client provides methods for authenticating with the backend, making HTTP
    requests (GET, POST, DELETE), and verifying API keys. It is designed to simplify
    communication with the Tellurio platform.
    """

    def __init__(self, base_url: str = None, port: int = None):
        """
        Initializes the TellurioClient instance.

        Args:
            base_url (str, optional): The base URL of the Tellurio backend. If not
                provided, it defaults to the value of the
                `TELLURIO_BACKEND_HTTP_BASE_URL` environment variable
                or "https://platform.tellurio.ai".
            port (int, optional): The port number for the backend. If not provided,
                it defaults to the value of the `TELLURIO_BACKEND_HTTP_PORT`
                environment variable or 443.
        """
        self.base_url = base_url or os.getenv(
            "TELLURIO_BACKEND_HTTP_BASE_URL", "https://platform.tellurio.ai"
        )
        self.port = port or os.getenv("TELLURIO_BACKEND_HTTP_PORT", 443)
        self.url = f"{self.base_url}:{self.port}"
        self.service_name = os.getenv(
            "KEYRING_SERVICE_NAME", "Tellurio"
        )  # Service name for keyring
        self.api_key = None

    def login(self, api_key: str = None, relogin: bool = False):
        """
        Logs in the user using an API key and verifies its validity.

        This method allows the user to provide an API key or retrieve a stored API key
        from the system. It verifies the API key by calling the backend and securely
        stores it using the `keyring` library if valid.

        Args:
            api_key (str, optional): The user's API key. If not provided, the method
                attempts to retrieve a stored API key from the local system.
            relogin (bool): If True, forces a re-login and requires the user to provide
                a new API key.

        Returns:
            dict: A dictionary containing the user's email and username.

        Raises:
            ValueError: If the API key is invalid or not provided during re-login.
        """
        # Use the provided API key if passed, otherwise check for stored key
        if api_key:
            self.api_key = api_key
        elif not relogin:
            username = load_username()
            self.api_key = keyring.get_password(self.service_name, username)
            if self.api_key:
                logger.info("Using stored API key from local keyring.")
            else:
                logger.error("No API key found in local keyring.")
                raise ValueError("API key is required for the first login.")
        else:
            logger.error("No API key provided for re-login.")
            raise ValueError("API key is required for re-login.")

        # Verify the API key
        response_data = self._verify_api_key()
        if response_data:
            email = response_data.get("email", "unknown user")
            username = response_data.get("username", "unknown user")
            logger.debug(f"API key is valid for user '{username}'.")

            # Save the API key securely only if it was provided and is valid
            if api_key:
                keyring.set_password(
                    self.service_name, response_data["username"], self.api_key
                )
                logger.info("API key provided and stored securely in local keyring.")
                save_username(response_data["username"])

            return {
                "email": email,
                "username": username,
            }
        else:
            logger.warning("Invalid API key. Please provide a valid API key.")
            if relogin:
                raise InvalidAPIKeyError("Re-login failed due to invalid API key.")
            raise InvalidAPIKeyError("Login failed due to invalid API key.")

    def get(self, endpoint: str) -> httpx.Response:
        """
        Makes a GET request to the specified endpoint.

        Args:
            endpoint (str): The API endpoint (relative to the base URL).

        Returns:
            httpx.Response: The HTTP response object.
        """
        url = f"{self.url}{endpoint}"
        headers = {
            "Authorization": f"Api-Key {self.api_key}",
            "Accept": "*/*",
        }

        try:
            with httpx.Client() as client:
                response = client.get(url, headers=headers)
            return response
        except httpx.RequestError as e:
            logger.error(f"Network error occurred while making GET request: {e}")
            raise ValueError("Network error occurred. Please check your connection.")
        except Exception as e:
            logger.error(f"An unexpected error occurred: {e}")
            raise ValueError("An unexpected error occurred. Please try again later.")

    def post(self, endpoint: str, json: dict) -> httpx.Response:
        """
        Makes a POST request to the specified endpoint.

        Args:
            endpoint (str): The API endpoint (relative to the base URL).
            json (dict): The JSON payload to send in the request.

        Returns:
            httpx.Response: The HTTP response object.
        """
        url = f"{self.url}{endpoint}"
        headers = {
            "Authorization": f"Api-Key {self.api_key}",
            "Content-Type": "application/json",
        }

        try:
            with httpx.Client() as client:
                response = client.post(url, headers=headers, json=json)
            return response
        except httpx.RequestError as e:
            logger.error(f"Network error occurred while making POST request: {e}")
            raise ValueError("Network error occurred. Please check your connection.")
        except Exception as e:
            logger.error(f"An unexpected error occurred: {e}")
            raise ValueError("An unexpected error occurred. Please try again later.")

    def patch(self, endpoint: str, json: dict) -> httpx.Response:
        """
        Makes a PATCH request to the specified endpoint.

        Args:
            endpoint (str): The API endpoint (relative to the base URL).
            json (dict): The JSON payload to send in the request.

        Returns:
            httpx.Response: The HTTP response object.
        """
        url = f"{self.url}{endpoint}"
        headers = {
            "Authorization": f"Api-Key {self.api_key}",
            "Content-Type": "application/json",
        }

        try:
            with httpx.Client() as client:
                response = client.patch(url, headers=headers, json=json)
            return response
        except httpx.RequestError as e:
            logger.error(f"Network error occurred while making PATCH request: {e}")
            raise ValueError("Network error occurred. Please check your connection.")
        except Exception as e:
            logger.error(f"An unexpected error occurred: {e}")
            raise ValueError("An unexpected error occurred. Please try again later.")

    def delete(self, endpoint: str) -> httpx.Response:
        """
        Makes a DELETE request to the specified endpoint.

        Args:
            endpoint (str): The API endpoint (relative to the base URL).

        Returns:
            httpx.Response: The HTTP response object.
        """
        url = f"{self.url}{endpoint}"
        headers = {
            "Authorization": f"Api-Key {self.api_key}",
            "Accept": "*/*",
        }

        try:
            with httpx.Client() as client:
                response = client.delete(url, headers=headers)
            return response
        except httpx.RequestError as e:
            logger.error(f"Network error occurred while making DELETE request: {e}")
            raise ValueError("Network error occurred. Please check your connection.")
        except Exception as e:
            logger.error(f"An unexpected error occurred: {e}")
            raise ValueError("An unexpected error occurred. Please try again later.")

    def _verify_api_key(self) -> dict:
        """
        Verifies the validity of the API key
        by calling the /api/v0/verify-api-key/ endpoint.

        Returns:
            dict: A dictionary containing the user's email, username and a message
            indicating if the API key is valid, None otherwise.
        """
        endpoint = "/api/v0/verify-api-key/"
        try:
            response = self.get(endpoint)

            if response.status_code == 200:
                try:
                    data = response.json()
                    logger.debug(f"API key verification successful: {data}")
                    return data
                except ValueError:
                    logger.error("Failed to parse JSON response from backend.")
                    return None
            elif response.status_code == 401:
                logger.warning("API key is invalid or missing.")
            else:
                logger.error(f"Error: {response.status_code} - {response.text}")
        except ValueError as e:
            logger.error(f"Error during API key verification: {e}")
            raise

        return None


def get_default_client() -> tuple[TellurioClient, TellurioWebSocketClient]:
    """
    Returns the global default TellurioClient instance.

    This function initializes a global instance of the TellurioClient if it does not
    already exist and returns it. The global instance can be used as a singleton for
    interacting with the backend.

    Returns:
        TellurioClient: The global default TellurioClient instance.
    """
    global _default_client, _default_ws_client

    if _default_client is None:
        # Initialize the default HTTP client
        _default_client = TellurioClient()

    if _default_ws_client is None:
        # Initialize default WebSocket client
        _default_ws_client = TellurioWebSocketClient()

    return _default_client, _default_ws_client



import logging
from abc import ABC
from typing import Dict, List, Optional, Union

from afnio.logging_config import configure_logging
from afnio.tellurio._eventloop import run_in_background_loop
from afnio.tellurio._model_registry import register_model
from afnio.tellurio.client import get_default_client
from afnio.tellurio.consent import check_consent

# Configure logging
configure_logging()
logger = logging.getLogger(__name__)


class BaseModel(ABC):
    """
    An abstraction for a model.
    """

    def __init__(
        self,
        provider: str = None,
        config: Optional[dict] = None,
        usage: Optional[dict] = None,
    ):
        self.provider = provider
        self._config = config or {}
        self._usage = usage or {}
        self.model_id = None

        # Request user consent before sending sensitive info to the server
        check_consent()

        try:
            # Get the singleton websocket client
            _, ws_client = get_default_client()

            payload = {
                "class_type": self.__class__.__name__,
                "provider": self.provider,
                "config": self.get_config(),
                "usage": self.get_usage(),
            }
            response = run_in_background_loop(ws_client.call("create_model", payload))
            logger.debug(f"LM model created and shared with the server: {self!r}")
            model_id = response["result"].get("model_id")
            if not model_id:
                logger.error(
                    f"Server did not return a model_id "
                    f"for payload: {payload!r}, response: {response!r}"
                )
                raise RuntimeError(
                    "Failed to create LM model: server did not return a model_id."
                )
            self.model_id = model_id
            register_model(self)
        except Exception as e:
            logger.error(f"Failed to share LM model with the server: {e}")
            raise

    def get_provider(self) -> Optional[str]:
        """Returns the model provider name."""
        return self.provider

    def get_config(self) -> Dict[str, Union[str, float, int]]:
        """
        Returns the model configuration.
        This includes the model name, temperature, max tokens, and other
        parameters that are used to configure the model's behavior.

        Returns:
            dict: A dictionary containing the model's configuration parameters.
        """
        return self._config

    def update_usage(self, usage: Dict[str, int]) -> None:
        """
        Updates the internal token usage statistics.

        Each model provider (e.g., OpenAI, Anthropic) may have a different usage format.
        This method should be implemented by subclasses to ensure correct parsing
        and aggregation of token usage.

        Args:
            usage (Dict[str, int]): A dictionary containing token usage metrics,
                such as `prompt_tokens`, `completion_tokens`, and `total_tokens`.

        Raises:
            NotImplementedError: If called on the base class without an implementation.
        """
        raise NotImplementedError

    def get_usage(self) -> Dict[str, int]:
        """
        Retrieves the current token usage statistics.

        Returns:
            Dict[str, int]: A dictionary containing cumulative token usage
                statistics since the model instance was initialized.

        Example:
            >>> model.get_usage()
            {'prompt_tokens': 1500, 'completion_tokens': 1200, 'total_tokens': 2700}
        """
        return self._usage.copy()

    def clear_usage(self) -> None:
        """
        Clears the token usage statistics.

        This resets all numerical values in the usage dictionary to zero (including
        nested values), while preserving the dictionary structure.
        """

        def zero_out_values(
            d: Dict[str, Union[int, Dict[str, int]]],
        ) -> Dict[str, Union[int, Dict[str, int]]]:
            return {
                key: zero_out_values(value) if isinstance(value, dict) else 0
                for key, value in d.items()
            }

        if hasattr(self, "_usage"):
            self._usage = zero_out_values(self._usage)
        else:
            self._usage = {}

    def __deepcopy__(self, memo):
        """
        Custom deepcopy to save only the class type and metadata like usage.
        """
        if id(self) in memo:
            return memo[id(self)]
        # Save only the class type and any necessary metadata (e.g., usage details)
        cls_copy = {
            "class_type": self.__class__.__name__,
            "provider": self.provider,
            "usage": self.get_usage(),
        }

        # Store the copied object in memo before returning it
        memo[id(self)] = cls_copy
        return cls_copy


# TODO: handle caching
class TextCompletionModel(BaseModel):
    """
    An abstraction for a language model that accepts a prompt composed of a single
    text input and generates a textual completion.
    """

    def __init__(self, provider: str = None, **kwargs):
        super().__init__(provider=provider, **kwargs)

    async def acomplete(self, prompt: str, **kwargs) -> str:
        """
        Asynchronous method to generate a completion for the given prompt.

        Args:
            prompt (str): The input text for which the model should generate
                a completion.
            **kwargs: Additional parameters to configure the model's behavior during
                chat completion. This may include options such as:
                - model (str): The model to use (e.g., "gpt-4o").
                - temperature (float): Amount of randomness injected into the response.
                - max_completion_tokens (int): Maximum number of tokens to generate.
                - etc.

                For a complete list of supported parameters for each model, refer to the
                respective API documentation.

        Returns:
            str: A string containing the generated completion.
        """
        raise NotImplementedError

    def complete(self, prompt: str, **kwargs) -> str:
        """
        Synchronous method to generate a completion for the given prompt.

        Args:
            prompt (str): The input text for which the model should generate
                a completion.
            **kwargs: Additional parameters to configure the model's behavior during
                chat completion. This may include options such as:
                - model (str): The model to use (e.g., "gpt-4o").
                - temperature (float): Amount of randomness injected into the response.
                - max_completion_tokens (int): Maximum number of tokens to generate.
                - etc.

                For a complete list of supported parameters for each model, refer to the
                respective API documentation.

        Returns:
            str: A string containing the generated completion.
        """
        raise NotImplementedError


# TODO: handle caching
class ChatCompletionModel(BaseModel):
    """
    An abstraction for a language model that accepts a prompt composed of an array
    of messages containing instructions for the model. Each message can have a
    different role, influencing how the model interprets the input.
    """

    def __init__(self, provider: str = None, **kwargs):
        super().__init__(provider=provider, **kwargs)

    # TODO: Add link to `API documentation` for kwargs of each supported model
    async def achat(self, messages: List[Dict[str, str]], **kwargs) -> str:
        """
        Asynchronous method to handle chat-based interactions with the model.

        Args:
            messages (List[Dict[str, str]]): A list of messages, where each message
                is represented as a dictionary with "role" (e.g., "user", "system")
                and "content" (the text of the message).
            **kwargs: Additional parameters to configure the model's behavior during
                chat completion. This may include options such as:
                - model (str): The model to use (e.g., "gpt-4o").
                - temperature (float): Amount of randomness injected into the response.
                - max_completion_tokens (int): Maximum number of tokens to generate.
                - etc.

                For a complete list of supported parameters for each model, refer to the
                respective API documentation.

        Returns:
            str: A string containing the model's response to the chat messages.
        """
        raise NotImplementedError

    def chat(self, messages: List[Dict[str, str]], **kwargs) -> str:
        """
        Synchronous method to handle chat-based interactions with the model.

        Args:
            messages (List[Dict[str, str]]): A list of messages, where each message
                is represented as a dictionary with "role" (e.g., "user", "system")
                and "content" (the text of the message).
            **kwargs: Additional parameters to configure the model's behavior during
                chat completion. This may include options such as:
                - model (str): The model to use (e.g., "gpt-4o").
                - temperature (float): Amount of randomness injected into the response.
                - max_completion_tokens (int): Maximum number of tokens to generate.
                - etc.

                For a complete list of supported parameters for each model, refer to the
                respective API documentation.

        Returns:
            str: A string containing the model's response to the chat messages.
        """
        raise NotImplementedError


# TODO: handle caching
class EmbeddingModel(BaseModel):
    """
    An abstraction for a model that generates embeddings for input texts.
    """

    def __init__(self, provider: str = None, **kwargs):
        super().__init__(provider=provider, **kwargs)

    async def aembed(self, input: List[str], **kwargs) -> List[List[float]]:
        """
        Asynchronous method to generate embeddings for the given input texts.

        Args:
            input (List[str]): A list of input strings for which embeddings
                should be generated.
            **kwargs: Additional parameters to configure the model's behavior during
                chat completion. This may include options such as:
                - model (str): The model to use (e.g., "gpt-4o").
                - temperature (float): Amount of randomness injected into the response.
                - max_completion_tokens (int): Maximum number of tokens to generate.
                - etc.

                For a complete list of supported parameters for each model, refer to the
                respective API documentation.

        Returns:
            List[List[float]]: A list of embeddings, where each embedding is represented
                as a list of floats corresponding to the input strings.
        """
        raise NotImplementedError

    def embed(self, input: List[str], **kwargs) -> List[List[float]]:
        """
        Synchronous method to generate embeddings for the given input texts.

        Args:
            input (List[str]): A list of input strings for which embeddings
                should be generated.
            **kwargs: Additional parameters to configure the model's behavior during
                chat completion. This may include options such as:
                - model (str): The model to use (e.g., "gpt-4o").
                - temperature (float): Amount of randomness injected into the response.
                - max_completion_tokens (int): Maximum number of tokens to generate.
                - etc.

                For a complete list of supported parameters for each model, refer to the
                respective API documentation.

        Returns:
            List[List[float]]: A list of embeddings, where each embedding is represented
                as a list of floats corresponding to the input strings.
        """
        raise NotImplementedError



import os

import pytest

from afnio._variable import Variable
from afnio.autodiff.basic_ops import Add, Split
from afnio.autodiff.evaluator import DeterministicEvaluator
from afnio.autodiff.grad_mode import no_grad
from afnio.tellurio import login
from afnio.tellurio.run import init


@pytest.fixture(autouse=True)
def login_and_ensure_default_run():
    """
    Test the login function with real HTTP and WebSocket connections and
    ensure a default Run exists and is set as active before tests.
    """
    # Log in to the Tellurio service using the API key
    api_key = os.getenv("TEST_ACCOUNT_API_KEY", "valid_api_key")
    login(api_key=api_key)

    # Use your test org/project names from env or defaults
    namespace_slug = os.getenv("TEST_ORG_SLUG", "tellurio-test")
    project_display_name = os.getenv("TEST_PROJECT", "Test Project")
    run = init(namespace_slug, project_display_name)
    return run


@pytest.fixture
def variables():
    x = Variable(data="abc", role="first input", requires_grad=True)
    y = Variable(data="def", role="second input", requires_grad=False)
    return x, y


class TestFunctionSync:
    """
    Tests for forward and backward passes of Function operations in afnio.autodiff.
    """

    def test_forward_add(self, variables):
        """
        Test the Add function's forward pass with two Variable inputs.
        """
        x, y = variables
        result = Add.apply(x, y)
        assert isinstance(result, Variable)
        assert result.data == "abcdef"
        assert result.role == "first input and second input"
        assert result.requires_grad is True
        assert "<AddBackward" in str(result.grad_fn)
        assert "<afnio.autodiff.function.AddBackward" in repr(result.grad_fn)
        assert len(result.grad_fn.next_functions) == 2
        assert "<AccumulateGrad" in str(result.grad_fn.next_functions[0].node)
        result.grad_fn.next_functions[0].output_nr == 0
        assert "None" in str(result.grad_fn.next_functions[1].node)
        result.grad_fn.next_functions[1].output_nr == 0
        assert result.is_leaf is False

    def test_forward_add_no_grad(self, variables):
        """
        Test the Add function's forward pass with two Variable inputs
        and using the `no_grad()` context manager.
        """
        x, y = variables
        with no_grad():
            result = Add.apply(x, y)
        assert isinstance(result, Variable)
        assert result.data == "abcdef"
        assert result.role == "first input and second input"
        assert result.requires_grad is False
        assert result.grad_fn is None
        assert result.is_leaf is True

    def test_forward_split(self, variables):
        """
        Test the Split function's forward pass with single Variable input.
        """
        x, _ = variables
        x.data = "a b c"
        result = Split.apply(x, sep=" ")
        assert isinstance(result, tuple)
        assert all(isinstance(v, Variable) for v in result)
        assert [v.data for v in result] == ["a", "b", "c"]
        expected_roles = [f"split part {i} of first input" for i in range(len(result))]
        for i, v in enumerate(result):
            assert v.role == expected_roles[i]
            assert v.requires_grad is True
            assert "<SplitBackward" in str(result[i].grad_fn)
            assert "<afnio.autodiff.function.SplitBackward" in repr(result[i].grad_fn)
            assert len(v.grad_fn.next_functions) == 1
            assert "<AccumulateGrad" in str(v.grad_fn.next_functions[0].node)
            assert v.grad_fn.next_functions[0].output_nr == 0

    def test_forward_deterministic_evaluator(self):
        """
        Test the DeterministicEvaluator function's forward pass with Callable input.
        """

        def exact_match_fn(pred: str, tgt: str) -> int:
            return 1 if pred == tgt else 0

        fn_purpose = "exact match"
        prediction = Variable(data="green", role="color prediction", requires_grad=True)
        target = Variable(data="red", role="expected color")
        result = DeterministicEvaluator.apply(
            prediction, target, exact_match_fn, fn_purpose, None, None, None
        )
        score, explanation = result
        assert isinstance(result, tuple)
        assert isinstance(score, Variable)
        assert isinstance(explanation, Variable)

        # Check score and explanation attributes
        assert score.data == 0
        assert score.role == "Evaluation result score of color prediction"
        assert score.requires_grad is True
        assert explanation.data == (
            "The evaluation function, designed for 'exact match', "
            "compared the <DATA> field of the predicted variable ('green') with "
            "the <DATA> field of the target variable ('red'), "
            "resulting in a score: 0."
        )
        assert explanation.role == "Evaluation result explanation of color prediction"
        assert explanation.requires_grad is True

        # Check grad_fn for score and explanation
        for var in (score, explanation):
            assert "<DeterministicEvaluatorBackward" in str(var.grad_fn)
            assert len(var.grad_fn.next_functions) == 2
            assert "<AccumulateGrad" in str(var.grad_fn.next_functions[0].node)
            assert var.grad_fn.next_functions[0].output_nr == 0
            assert "None" in str(var.grad_fn.next_functions[1].node)
            assert var.grad_fn.next_functions[1].output_nr == 0

    def test_backward_add(self, variables):
        """
        Test the Add function's backward pass.
        """
        x, y = variables
        result = Add.apply(x, y)
        assert isinstance(result, Variable)
        assert result.data == "abcdef"

        gradient = Variable(data="MY_FEEDBACK", role="add gradient")
        result.backward(gradient)

        assert len(x.grad) == 1
        assert y.grad == []  # requires_grad=False, so no gradient
        assert x.grad[0].data == (
            "Here is the combined feedback we got for this specific "
            "first input and other variables: MY_FEEDBACK"
        )
        assert x.grad[0].role == "feedback to first input"

    def test_backward_clear_pending_grad(self):
        """
        Test that _pending_grad is set during backward
        and cleared after 'clear_backward'.

        DAG:
                b → c → d
              ↗           ↘
            a  → → → → → → e
        """
        a = Variable(data="abc_", role="first input", requires_grad=True)
        b = a + Variable(data="def_", role="second input")
        c = b + Variable(data="ghi_", role="third input")
        d = c + Variable(data="jkl_", role="fourth input")
        e = Add.apply(a, d)

        gradient = Variable(data="MY_FEEDBACK", role="add gradient")
        e.backward(gradient)

        # The variable should have _pending_grad set before clear_backward
        assert a._pending_grad is True

        # We are only able to read `a.grad` when we exit `_wait_for_pending()`
        # and at that point, _pending_grad should be True
        assert len(a.grad) == 2
        assert a._pending_grad is False



import threading
from contextlib import contextmanager

# Thread-local flag to control gradient tracking
_grad_enabled = threading.local()
_grad_enabled.enabled = True  # By default, gradients are enabled


def is_grad_enabled() -> bool:
    """Check whether gradients are currently enabled."""
    return getattr(_grad_enabled, "enabled", True)


def set_grad_enabled(mode: bool):
    """Set the global state of gradient tracking."""
    _grad_enabled.enabled = mode


@contextmanager
def no_grad():
    """
    Context manager that disables gradient calculation. All operations within this block
    will not track gradients, making them more memory-efficient.

    Disabling gradient calculation is useful for inference, when you are sure
    that you will not call :meth:`Variable.backward()`. It will reduce memory
    consumption for computations that would otherwise have `requires_grad=True`.

    In this mode, the result of every computation will have
    `requires_grad=False`, even when the inputs have `requires_grad=True`.
    There is an exception! All factory functions, or functions that create
    a new Variable and take a requires_grad kwarg, will NOT be affected by
    this mode.

    This context manager is thread local; it will not affect computation
    in other threads.

    Also functions as a decorator.

    Example::
        >>> x = hf.Variable("abc", role="variable", requires_grad=True)
        >>> with hf.no_grad():
        ...     y = x + x
        >>> y.requires_grad
        False
        >>> @hf.no_grad()
        ... def doubler(x):
        ...     return x + x
        >>> z = doubler(x)
        >>> z.requires_grad
        False
        >>> @hf.no_grad
        ... def tripler(x):
        ...     return x + x + x
        >>> z = tripler(x)
        >>> z.requires_grad
        False
        >>> # factory function exception
        >>> with hf.no_grad():
        ...     a = hf.cognitive.Parameter("xyz")
        >>> a.requires_grad
        True
    """
    previous_state = is_grad_enabled()  # Store the current state
    set_grad_enabled(False)  # Disable gradients
    try:
        yield  # Execute the block
    finally:
        set_grad_enabled(previous_state)  # Restore the original state



import copy
import json
import os
from typing import Dict, Iterable, List, Mapping, Optional, Union

import httpx
from openai import DEFAULT_MAX_RETRIES, NOT_GIVEN, NotGiven
from openai import AsyncOpenAI as AsyncOpenAICli
from openai import OpenAI as OpenAICli
from openai.types.chat import (
    ChatCompletionReasoningEffort,
    completion_create_params,
)
from openai.types.chat.chat_completion_message_param import ChatCompletionMessageParam
from openai.types.chat.chat_completion_prediction_content_param import (
    ChatCompletionPredictionContentParam,
)
from openai.types.chat.chat_completion_stream_options_param import (
    ChatCompletionStreamOptionsParam,
)
from openai.types.chat.chat_completion_tool_choice_option_param import (
    ChatCompletionToolChoiceOptionParam,
)
from openai.types.chat.chat_completion_tool_param import ChatCompletionToolParam
from openai.types.chat_model import ChatModel
from openai.types.completion_usage import CompletionUsage
from typing_extensions import Literal

from .model import ChatCompletionModel, EmbeddingModel, TextCompletionModel

PROVIDER = "openai"
INITIAL_USAGE = {
    "completion_tokens": 0,
    "prompt_tokens": 0,
    "total_tokens": 0,
    "prompt_tokens_details": {
        "cached_tokens": 0,
        "audio_tokens": 0,
    },
    "completion_tokens_details": {
        "reasoning_tokens": 0,
        "audio_tokens": 0,
        "accepted_prediction_tokens": 0,
        "rejected_prediction_tokens": 0,
    },
}


class Omit:
    """In certain situations you need to be able to represent a case where a default value has
    to be explicitly removed and `None` is not an appropriate substitute, for example:

    ```py
    # as the default `Content-Type` header is `application/json` that will be sent
    client.post("/upload/files", files={"file": b"my raw file content"})

    # you can't explicitly override the header as it has to be dynamically generated
    # to look something like: 'multipart/form-data; boundary=0d8382fcf5f8c3be01ca2e11002d2983'
    client.post(..., headers={"Content-Type": "multipart/form-data"})

    # instead you can remove the default `application/json` header by passing Omit
    client.post(..., headers={"Content-Type": Omit()})
    ```
    """  # noqa: E501

    def __bool__(self) -> Literal[False]:
        return False


Headers = Mapping[str, Union[str, Omit]]
Query = Mapping[str, object]
Body = object


class OpenAI(
    TextCompletionModel,
    ChatCompletionModel,
    EmbeddingModel,
    OpenAICli,
):
    """
    OpenAI synchronous client to perform multiple language model operations.
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        organization: Optional[str] = None,
        project: Optional[str] = None,
        base_url: Optional[Union[str, httpx.URL]] = None,
        websocket_base_url: Optional[Union[str, httpx.URL]] = None,
        timeout: Union[float, httpx.Timeout, None, NotGiven] = NOT_GIVEN,
        max_retries: int = DEFAULT_MAX_RETRIES,
        default_headers: Optional[Mapping[str, str]] = None,
        default_query: Optional[Mapping[str, object]] = None,
        http_client: Optional[httpx.Client] = None,
    ):
        usage = copy.deepcopy(INITIAL_USAGE)  # Ensure a fresh copy

        # Validate and build config
        config = {
            "api_key": api_key or os.getenv("OPENAI_API_KEY"),
            "organization": organization,
            "project": project,
            "base_url": base_url,
            "websocket_base_url": websocket_base_url,
            "timeout": timeout,
            "max_retries": max_retries,
            "default_headers": default_headers,
            "default_query": default_query,
        }
        # Remove None and NOT_GIVEN values
        config = {
            k: v for k, v in config.items() if v is not None and v is not NOT_GIVEN
        }
        # Validate serializability
        for k, v in config.items():
            _validate_config_param(k, v)

        self._client = OpenAICli(
            api_key=api_key or os.getenv("OPENAI_API_KEY"),
            organization=organization,
            project=project,
            base_url=base_url,
            websocket_base_url=websocket_base_url,
            timeout=timeout,
            max_retries=max_retries,
            default_headers=default_headers,
            default_query=default_query,
            http_client=http_client,
        )
        super().__init__(provider=PROVIDER, config=config, usage=usage)

    # TODO: Finalize implementation
    def complete(self, prompt: str) -> str:
        """
        Synchronous method to generate a completion for the given prompt.

        Args:
            prompt (str): The input text for which the model should generate
              a completion.

        Returns:
            str: A string containing the generated completion.
        """
        raise NotImplementedError

    def chat(
        self,
        *,
        messages: Iterable[ChatCompletionMessageParam],
        model: Union[str, ChatModel],
        frequency_penalty: Union[Optional[float], NotGiven] = NOT_GIVEN,
        logit_bias: Union[Optional[Dict[str, int]], NotGiven] = NOT_GIVEN,
        logprobs: Union[Optional[bool], NotGiven] = NOT_GIVEN,
        max_completion_tokens: Union[Optional[int], NotGiven] = NOT_GIVEN,
        metadata: Union[Optional[Dict[str, str]], NotGiven] = NOT_GIVEN,
        n: Union[Optional[int], NotGiven] = NOT_GIVEN,
        parallel_tool_calls: Union[bool, NotGiven] = NOT_GIVEN,
        prediction: Union[
            Optional[ChatCompletionPredictionContentParam], NotGiven
        ] = NOT_GIVEN,
        presence_penalty: Union[Optional[float], NotGiven] = NOT_GIVEN,
        reasoning_effort: Union[ChatCompletionReasoningEffort, NotGiven] = NOT_GIVEN,
        response_format: Union[
            completion_create_params.ResponseFormat, NotGiven
        ] = NOT_GIVEN,
        seed: Union[Optional[int], NotGiven] = NOT_GIVEN,
        service_tier: Union[Optional[Literal["auto", "default"]], NotGiven] = NOT_GIVEN,
        stop: Union[Union[Optional[str], List[str]], NotGiven] = NOT_GIVEN,
        store: Union[Optional[bool], NotGiven] = NOT_GIVEN,
        # TODO: `stream` can be useful during inference, but forbid during training or backpropagation
        stream: Union[Optional[Literal[False]], NotGiven] = NOT_GIVEN,
        stream_options: Union[
            Optional[ChatCompletionStreamOptionsParam], NotGiven
        ] = NOT_GIVEN,
        temperature: Union[Optional[float], NotGiven] = NOT_GIVEN,
        tool_choice: Union[ChatCompletionToolChoiceOptionParam, NotGiven] = NOT_GIVEN,
        tools: Union[Iterable[ChatCompletionToolParam], NotGiven] = NOT_GIVEN,
        top_logprobs: Union[Optional[int], NotGiven] = NOT_GIVEN,
        top_p: Union[Optional[float], NotGiven] = NOT_GIVEN,
        user: Union[str, NotGiven] = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Optional[Headers] = None,
        extra_query: Optional[Query] = None,
        extra_body: Optional[Body] = None,
        timeout: Optional[Union[float, httpx.Timeout, NotGiven]] = NOT_GIVEN,
    ) -> str:
        """Synchronously creates a model response for the given chat conversation.

        Learn more in the
        [text generation](https://platform.openai.com/docs/guides/text-generation)
        and [chat](https://platform.openai.com/docs/api-reference/chat) guides.

        Parameter support can differ depending on the model used to generate the
        response, particularly for newer reasoning models.

        Args:
          messages: A list of messages comprising the conversation so far. Depending on
            the [model](https://platform.openai.com/docs/models) you use, different
            message types (modalities) are supported. Refer to
            [text](https://platform.openai.com/docs/guides/text-generation).

          model: ID of the model to use. See the
            [model endpoint compatibility](https://platform.openai.com/docs/models#model-endpoint-compatibility)
            table for details on which models work with the Chat API.

          frequency_penalty: Number between -2.0 and 2.0. Positive values penalize new
            tokens based on their existing frequency in the text so far, decreasing the
            model's likelihood to repeat the same line verbatim.

          logit_bias: Modify the likelihood of specified tokens appearing in the
            completion.

            Accepts a JSON object that maps tokens (specified by their token ID in the
            tokenizer) to an associated bias value from -100 to 100. Mathematically, the
            bias is added to the logits generated by the model prior to sampling. The
            exact effect will vary per model, but values between -1 and 1 should
            decrease or increase likelihood of selection; values like -100 or 100 should
            result in a ban or exclusive selection of the relevant token.

          logprobs: Whether to return log probabilities of the output tokens or not. If
            true, returns the log probabilities of each output token returned in the
            `content` of `message`.

          max_completion_tokens: An upper bound for the number of tokens that can be
            generated for a completion, including visible output tokens and
            [reasoning tokens](https://platform.openai.com/docs/guides/reasoning).

          metadata: Developer-defined tags and values used for filtering completions in
            the [dashboard](https://platform.openai.com/chat-completions).

          n: How many chat completion choices to generate for each input message. Note
            that you will be charged based on the number of generated tokens across all
            of the choices. Keep `n` as `1` to minimize costs.

          parallel_tool_calls: Whether to enable
            [parallel function calling](https://platform.openai.com/docs/guides/function-calling#configuring-parallel-function-calling)
            during tool use.

          prediction: Static predicted output content, such as the content of a text
            file that is being regenerated.

          presence_penalty: Number between -2.0 and 2.0. Positive values penalize new
            tokens based on whether they appear in the text so far, increasing the
            model's likelihood to talk about new topics.

          reasoning_effort: **o1 models only**

            Constrains effort on reasoning for
            [reasoning models](https://platform.openai.com/docs/guides/reasoning).
            Currently supported values are `low`, `medium`, and `high`. Reducing
            reasoning effort can result in faster responses and fewer tokens used on
            reasoning in a response.

          response_format: An object specifying the format that the model must output.

            Setting to `{ "type": "json_schema", "json_schema": {...} }` enables
            Structured Outputs which ensures the model will match your supplied
            JSON schema. Learn more in the
            [Structured Outputs guide](https://platform.openai.com/docs/guides/structured-outputs).

            Setting to `{ "type": "json_object" }` enables JSON mode, which ensures the
            message the model generates is valid JSON.

            **Important:** when using JSON mode, you **must** also instruct the model to
            produce JSON yourself via a system or user message. Without this, the model
            may generate an unending stream of whitespace until the generation reaches
            the token limit, resulting in a long-running and seemingly "stuck" request.
            Also note that the message content may be partially cut off if
            `finish_reason="length"`, which indicates the generation exceeded
            `max_tokens` or the conversation exceeded the max context length.

          seed: This feature is in Beta. If specified, our system will make a best
            effort to sample deterministically, such that repeated requests with the
            same `seed` and parameters should return the same result. Determinism is not
            guaranteed, and you should refer to the `system_fingerprint` response
            parameter to monitor changes in the backend.

          service_tier: Specifies the latency tier to use for processing the request.
            This parameter is relevant for customers subscribed to the scale tier
            service:

            - If set to 'auto', and the Project is Scale tier enabled, the system will
              utilize scale tier credits until they are exhausted.
            - If set to 'auto', and the Project is not Scale tier enabled, the request
              will be processed using the default service tier with a lower uptime SLA
              and no latency guarentee.
            - If set to 'default', the request will be processed using the default
              service tier with a lower uptime SLA and no latency guarentee.
            - When not set, the default behavior is 'auto'.

            When this parameter is set, the response body will include the
            `service_tier` utilized.

          stop: Up to 4 sequences where the API will stop generating further tokens.

          store: Whether or not to store the output of this chat completion request for
            use in our
            [model distillation](https://platform.openai.com/docs/guides/distillation)
            or [evals](https://platform.openai.com/docs/guides/evals) products.

          stream: If set, partial message deltas will be sent, like in ChatGPT. Tokens
            will be sent as data-only
            [server-sent events](https://developer.mozilla.org/en-US/docs/Web/API/Server-sent_events/Using_server-sent_events#Event_stream_format)
            as they become available, with the stream terminated by a `data: [DONE]`
            message.
            [Example Python code](https://cookbook.openai.com/examples/how_to_stream_completions).

          stream_options: Options for streaming response. Only set this when you set
            `stream: true`.

          temperature: What sampling temperature to use, between 0 and 2. Higher values
            like 0.8 will make the output more random, while lower values like 0.2 will
            make it more focused and deterministic. We generally recommend altering this
            or `top_p` but not both.

          tool_choice: Controls which (if any) tool is called by the model. `none` means
            the model will not call any tool and instead generates a message. `auto`
            means the model can pick between generating a message or calling one or more
            tools. `required` means the model must call one or more tools. Specifying a
            particular tool via
            `{"type": "function", "function": {"name": "my_function"}}` forces the model
            to call that tool.

            `none` is the default when no tools are present. `auto` is the default if
            tools are present.

          tools: A list of tools the model may call. Currently, only functions are
            supported as a tool. Use this to provide a list of functions the model may
            generate JSON inputs for. A max of 128 functions are supported.

          top_logprobs: An integer between 0 and 20 specifying the number of most likely
            tokens to return at each token position, each with an associated log
            probability. `logprobs` must be set to `true` if this parameter is used.

          top_p: An alternative to sampling with temperature, called nucleus sampling,
            where the model considers the results of the tokens with top_p probability
            mass. So 0.1 means only the tokens comprising the top 10% probability mass
            are considered.

            We generally recommend altering this or `temperature` but not both.

          user: A unique identifier representing your end-user, which can help OpenAI to
            monitor and detect abuse.
            [Learn more](https://platform.openai.com/docs/guides/safety-best-practices#end-user-ids).

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in
            seconds
        """  # noqa: E501
        # TODO: handle `n > 1`` that returns multiple completions
        if isinstance(n, int) and n > 1:
            raise ValueError("n > 1 is not supported for chat completions.")

        response = self._client.chat.completions.create(
            messages=messages,
            model=model,
            frequency_penalty=frequency_penalty,
            logit_bias=logit_bias,
            logprobs=logprobs,
            max_completion_tokens=max_completion_tokens,
            metadata=metadata,
            n=n,
            parallel_tool_calls=parallel_tool_calls,
            prediction=prediction,
            presence_penalty=presence_penalty,
            reasoning_effort=reasoning_effort,
            response_format=response_format,
            seed=seed,
            service_tier=service_tier,
            stop=stop,
            store=store,
            stream=stream,
            stream_options=stream_options,
            temperature=temperature,
            tool_choice=tool_choice,
            tools=tools,
            top_logprobs=top_logprobs,
            top_p=top_p,
            user=user,
            extra_headers=extra_headers,
            extra_query=extra_query,
            extra_body=extra_body,
            timeout=timeout,
        )

        # Update usage statistics
        if hasattr(response, "usage") and response.usage:
            self.update_usage(response.usage)

        return response.choices[0].message.content

    # TODO: Finalize implementation
    def embed(self, input: List[str]) -> List[List[float]]:
        """
        Synchronous method to generate embeddings for the given input texts.

        Args:
            input (List[str]): A list of input strings for which embeddings
              should be generated.

        Returns:
            List[List[float]]: A list of embeddings, where each embedding is represented
              as a list of floats corresponding to the input strings.
        """
        raise NotImplementedError

    def update_usage(self, usage: CompletionUsage) -> None:
        """Updates the internal usage counters with values from a new API response.

        Args:
            usage (CompletionUsage): The usage object returned by the OpenAI API.
        """
        if not hasattr(self, "_usage"):
            self._usage = copy.deepcopy(INITIAL_USAGE)  # Ensure a fresh copy

        # Ensure we convert CompletionUsage to dict properly
        if isinstance(usage, CompletionUsage):
            usage = usage.model_dump()

        # Update core token usage fields
        self._usage["completion_tokens"] += usage.get("completion_tokens", 0)
        self._usage["prompt_tokens"] += usage.get("prompt_tokens", 0)
        self._usage["total_tokens"] += usage.get("total_tokens", 0)

        # Update prompt tokens details
        prompt_tokens_details = usage.get("prompt_tokens_details", {})
        self._usage["prompt_tokens_details"][
            "cached_tokens"
        ] += prompt_tokens_details.get("cached_tokens", 0)
        self._usage["prompt_tokens_details"][
            "audio_tokens"
        ] += prompt_tokens_details.get("audio_tokens", 0)

        # Update completion tokens details
        completion_tokens_details = usage.get("completion_tokens_details", {})
        self._usage["completion_tokens_details"][
            "reasoning_tokens"
        ] += completion_tokens_details.get("reasoning_tokens", 0)
        self._usage["completion_tokens_details"][
            "audio_tokens"
        ] += completion_tokens_details.get("audio_tokens", 0)
        self._usage["completion_tokens_details"][
            "accepted_prediction_tokens"
        ] += completion_tokens_details.get("accepted_prediction_tokens", 0)
        self._usage["completion_tokens_details"][
            "rejected_prediction_tokens"
        ] += completion_tokens_details.get("rejected_prediction_tokens", 0)


class AsyncOpenAI(
    TextCompletionModel,
    ChatCompletionModel,
    EmbeddingModel,
    AsyncOpenAICli,
):
    """
    OpenAI asynchronous client to perform multiple language model operations.
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        organization: Optional[str] = None,
        project: Optional[str] = None,
        base_url: Optional[Union[str, httpx.URL]] = None,
        websocket_base_url: Optional[Union[str, httpx.URL]] = None,
        timeout: Union[float, httpx.Timeout, None, NotGiven] = NOT_GIVEN,
        max_retries: int = DEFAULT_MAX_RETRIES,
        default_headers: Optional[Mapping[str, str]] = None,
        default_query: Optional[Mapping[str, object]] = None,
        http_client: Optional[httpx.AsyncClient] = None,
    ):
        usage = copy.deepcopy(INITIAL_USAGE)  # Ensure a fresh copy

        # Validate and build config
        config = {
            "api_key": api_key or os.getenv("OPENAI_API_KEY"),
            "organization": organization,
            "project": project,
            "base_url": base_url,
            "websocket_base_url": websocket_base_url,
            "timeout": timeout,
            "max_retries": max_retries,
            "default_headers": default_headers,
            "default_query": default_query,
        }
        # Remove None and NOT_GIVEN values
        config = {
            k: v for k, v in config.items() if v is not None and v is not NOT_GIVEN
        }
        # Validate serializability
        for k, v in config.items():
            _validate_config_param(k, v)

        self._aclient = AsyncOpenAICli(
            api_key=api_key or os.getenv("OPENAI_API_KEY"),
            organization=organization,
            project=project,
            base_url=base_url,
            websocket_base_url=websocket_base_url,
            timeout=timeout,
            max_retries=max_retries,
            default_headers=default_headers,
            default_query=default_query,
            http_client=http_client,
        )
        super().__init__(provider=PROVIDER, config=config, usage=usage)

    # TODO: Finalize implementation
    async def acomplete(self, prompt: str) -> str:
        """
        Asynchronous method to generate a completion for the given prompt.

        Args:
            prompt (str): The input text for which the model should generate
              a completion.

        Returns:
            str: A string containing the generated completion.
        """
        raise NotImplementedError

    async def achat(
        self,
        *,
        messages: Iterable[ChatCompletionMessageParam],
        model: Union[str, ChatModel],
        frequency_penalty: Union[Optional[float], NotGiven] = NOT_GIVEN,
        logit_bias: Union[Optional[Dict[str, int]], NotGiven] = NOT_GIVEN,
        logprobs: Union[Optional[bool], NotGiven] = NOT_GIVEN,
        max_completion_tokens: Union[Optional[int], NotGiven] = NOT_GIVEN,
        metadata: Union[Optional[Dict[str, str]], NotGiven] = NOT_GIVEN,
        n: Union[Optional[int], NotGiven] = NOT_GIVEN,
        parallel_tool_calls: Union[bool, NotGiven] = NOT_GIVEN,
        prediction: Union[
            Optional[ChatCompletionPredictionContentParam], NotGiven
        ] = NOT_GIVEN,
        presence_penalty: Union[Optional[float], NotGiven] = NOT_GIVEN,
        reasoning_effort: Union[ChatCompletionReasoningEffort, NotGiven] = NOT_GIVEN,
        response_format: Union[
            completion_create_params.ResponseFormat, NotGiven
        ] = NOT_GIVEN,
        seed: Union[Optional[int], NotGiven] = NOT_GIVEN,
        service_tier: Union[Optional[Literal["auto", "default"]], NotGiven] = NOT_GIVEN,
        stop: Union[Union[Optional[str], List[str]], NotGiven] = NOT_GIVEN,
        store: Union[Optional[bool], NotGiven] = NOT_GIVEN,
        # TODO: `stream` can be useful during inference, but forbid during training or backpropagation
        stream: Union[Optional[Literal[False]], NotGiven] = NOT_GIVEN,
        stream_options: Union[
            Optional[ChatCompletionStreamOptionsParam], NotGiven
        ] = NOT_GIVEN,
        temperature: Union[Optional[float], NotGiven] = NOT_GIVEN,
        tool_choice: Union[ChatCompletionToolChoiceOptionParam, NotGiven] = NOT_GIVEN,
        tools: Union[Iterable[ChatCompletionToolParam], NotGiven] = NOT_GIVEN,
        top_logprobs: Union[Optional[int], NotGiven] = NOT_GIVEN,
        top_p: Union[Optional[float], NotGiven] = NOT_GIVEN,
        user: Union[str, NotGiven] = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Optional[Headers] = None,
        extra_query: Optional[Query] = None,
        extra_body: Optional[Body] = None,
        timeout: Optional[Union[float, httpx.Timeout, NotGiven]] = NOT_GIVEN,
    ) -> str:
        """Asynchronously creates a model response for the given chat conversation.

        Learn more in the
        [text generation](https://platform.openai.com/docs/guides/text-generation)
        and [chat](https://platform.openai.com/docs/api-reference/chat) guides.

        Parameter support can differ depending on the model used to generate the
        response, particularly for newer reasoning models.

        Args:
          messages: A list of messages comprising the conversation so far. Depending on
            the [model](https://platform.openai.com/docs/models) you use, different
            message types (modalities) are supported. Refer to
            [text](https://platform.openai.com/docs/guides/text-generation).

          model: ID of the model to use. See the
            [model endpoint compatibility](https://platform.openai.com/docs/models#model-endpoint-compatibility)
            table for details on which models work with the Chat API.

          frequency_penalty: Number between -2.0 and 2.0. Positive values penalize new
            tokens based on their existing frequency in the text so far, decreasing the
            model's likelihood to repeat the same line verbatim.

          logit_bias: Modify the likelihood of specified tokens appearing in the
            completion.

            Accepts a JSON object that maps tokens (specified by their token ID in the
            tokenizer) to an associated bias value from -100 to 100. Mathematically, the
            bias is added to the logits generated by the model prior to sampling. The
            exact effect will vary per model, but values between -1 and 1 should
            decrease or increase likelihood of selection; values like -100 or 100 should
            result in a ban or exclusive selection of the relevant token.

          logprobs: Whether to return log probabilities of the output tokens or not. If
            true, returns the log probabilities of each output token returned in the
            `content` of `message`.

          max_completion_tokens: An upper bound for the number of tokens that can be
            generated for a completion, including visible output tokens and
            [reasoning tokens](https://platform.openai.com/docs/guides/reasoning).

          metadata: Developer-defined tags and values used for filtering completions in
            the [dashboard](https://platform.openai.com/chat-completions).

          n: How many chat completion choices to generate for each input message. Note
            that you will be charged based on the number of generated tokens across all
            of the choices. Keep `n` as `1` to minimize costs.

          parallel_tool_calls: Whether to enable
            [parallel function calling](https://platform.openai.com/docs/guides/function-calling#configuring-parallel-function-calling)
            during tool use.

          prediction: Static predicted output content, such as the content of a text
            file that is being regenerated.

          presence_penalty: Number between -2.0 and 2.0. Positive values penalize new
            tokens based on whether they appear in the text so far, increasing the
            model's likelihood to talk about new topics.

          reasoning_effort: **o1 models only**

            Constrains effort on reasoning for
            [reasoning models](https://platform.openai.com/docs/guides/reasoning).
            Currently supported values are `low`, `medium`, and `high`. Reducing
            reasoning effort can result in faster responses and fewer tokens used on
            reasoning in a response.

          response_format: An object specifying the format that the model must output.

            Setting to `{ "type": "json_schema", "json_schema": {...} }` enables
            Structured Outputs which ensures the model will match your supplied
            JSON schema. Learn more in the
            [Structured Outputs guide](https://platform.openai.com/docs/guides/structured-outputs).

            Setting to `{ "type": "json_object" }` enables JSON mode, which ensures the
            message the model generates is valid JSON.

            **Important:** when using JSON mode, you **must** also instruct the model to
            produce JSON yourself via a system or user message. Without this, the model
            may generate an unending stream of whitespace until the generation reaches
            the token limit, resulting in a long-running and seemingly "stuck" request.
            Also note that the message content may be partially cut off if
            `finish_reason="length"`, which indicates the generation exceeded
            `max_tokens` or the conversation exceeded the max context length.

          seed: This feature is in Beta. If specified, our system will make a best
            effort to sample deterministically, such that repeated requests with the
            same `seed` and parameters should return the same result. Determinism is not
            guaranteed, and you should refer to the `system_fingerprint` response
            parameter to monitor changes in the backend.

          service_tier: Specifies the latency tier to use for processing the request.
            This parameter is relevant for customers subscribed to the scale tier
            service:

            - If set to 'auto', and the Project is Scale tier enabled, the system will
              utilize scale tier credits until they are exhausted.
            - If set to 'auto', and the Project is not Scale tier enabled, the request
              will be processed using the default service tier with a lower uptime SLA
              and no latency guarentee.
            - If set to 'default', the request will be processed using the default
              service tier with a lower uptime SLA and no latency guarentee.
            - When not set, the default behavior is 'auto'.

            When this parameter is set, the response body will include the
            `service_tier` utilized.

          stop: Up to 4 sequences where the API will stop generating further tokens.

          store: Whether or not to store the output of this chat completion request for
            use in our
            [model distillation](https://platform.openai.com/docs/guides/distillation)
            or [evals](https://platform.openai.com/docs/guides/evals) products.

          stream: If set, partial message deltas will be sent, like in ChatGPT. Tokens
            will be sent as data-only
            [server-sent events](https://developer.mozilla.org/en-US/docs/Web/API/Server-sent_events/Using_server-sent_events#Event_stream_format)
            as they become available, with the stream terminated by a `data: [DONE]`
            message.
            [Example Python code](https://cookbook.openai.com/examples/how_to_stream_completions).

          stream_options: Options for streaming response. Only set this when you set
            `stream: true`.

          temperature: What sampling temperature to use, between 0 and 2. Higher values
            like 0.8 will make the output more random, while lower values like 0.2 will
            make it more focused and deterministic. We generally recommend altering this
            or `top_p` but not both.

          tool_choice: Controls which (if any) tool is called by the model. `none` means
            the model will not call any tool and instead generates a message. `auto`
            means the model can pick between generating a message or calling one or more
            tools. `required` means the model must call one or more tools. Specifying a
            particular tool via
            `{"type": "function", "function": {"name": "my_function"}}` forces the model
            to call that tool.

            `none` is the default when no tools are present. `auto` is the default if
            tools are present.

          tools: A list of tools the model may call. Currently, only functions are
            supported as a tool. Use this to provide a list of functions the model may
            generate JSON inputs for. A max of 128 functions are supported.

          top_logprobs: An integer between 0 and 20 specifying the number of most likely
            tokens to return at each token position, each with an associated log
            probability. `logprobs` must be set to `true` if this parameter is used.

          top_p: An alternative to sampling with temperature, called nucleus sampling,
            where the model considers the results of the tokens with top_p probability
            mass. So 0.1 means only the tokens comprising the top 10% probability mass
            are considered.

            We generally recommend altering this or `temperature` but not both.

          user: A unique identifier representing your end-user, which can help OpenAI to
            monitor and detect abuse.
            [Learn more](https://platform.openai.com/docs/guides/safety-best-practices#end-user-ids).

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in
            seconds
        """  # noqa: E501
        # TODO: handle `n > 1`` that returns multiple completions
        if isinstance(n, int) and n > 1:
            raise ValueError("n > 1 is not supported for async chat completions.")

        response = await self._aclient.chat.completions.create(
            messages=messages,
            model=model,
            frequency_penalty=frequency_penalty,
            logit_bias=logit_bias,
            logprobs=logprobs,
            max_completion_tokens=max_completion_tokens,
            metadata=metadata,
            n=n,
            parallel_tool_calls=parallel_tool_calls,
            prediction=prediction,
            presence_penalty=presence_penalty,
            reasoning_effort=reasoning_effort,
            response_format=response_format,
            seed=seed,
            service_tier=service_tier,
            stop=stop,
            store=store,
            stream=stream,
            stream_options=stream_options,
            temperature=temperature,
            tool_choice=tool_choice,
            tools=tools,
            top_logprobs=top_logprobs,
            top_p=top_p,
            user=user,
            extra_headers=extra_headers,
            extra_query=extra_query,
            extra_body=extra_body,
            timeout=timeout,
        )

        # Update usage statistics
        if hasattr(response, "usage") and response.usage:
            self.update_usage(response.usage)

        return response.choices[0].message.content

    # TODO: Finalize implementation
    async def aembed(self, input: List[str]) -> List[List[float]]:
        """
        Asynchronous method to generate embeddings for the given input texts.

        Args:
            input (List[str]): A list of input strings for which embeddings
              should be generated.

        Returns:
            List[List[float]]: A list of embeddings, where each embedding is represented
              as a list of floats corresponding to the input strings.
        """
        raise NotImplementedError

    def update_usage(self, usage: CompletionUsage) -> None:
        """Updates the internal usage counters with values from a new API response.

        Args:
            usage (CompletionUsage): The usage object returned by the OpenAI API.
        """
        if not hasattr(self, "_usage"):
            self._usage = copy.deepcopy(INITIAL_USAGE)  # Ensure a fresh copy

        # Ensure we convert CompletionUsage to dict properly
        if isinstance(usage, CompletionUsage):
            usage = usage.model_dump()

        # Update core token usage fields
        self._usage["completion_tokens"] += usage.get("completion_tokens", 0)
        self._usage["prompt_tokens"] += usage.get("prompt_tokens", 0)
        self._usage["total_tokens"] += usage.get("total_tokens", 0)

        # Update prompt tokens details
        prompt_tokens_details = usage.get("prompt_tokens_details", {})
        self._usage["prompt_tokens_details"][
            "cached_tokens"
        ] += prompt_tokens_details.get("cached_tokens", 0)
        self._usage["prompt_tokens_details"][
            "audio_tokens"
        ] += prompt_tokens_details.get("audio_tokens", 0)

        # Update completion tokens details
        completion_tokens_details = usage.get("completion_tokens_details", {})
        self._usage["completion_tokens_details"][
            "reasoning_tokens"
        ] += completion_tokens_details.get("reasoning_tokens", 0)
        self._usage["completion_tokens_details"][
            "audio_tokens"
        ] += completion_tokens_details.get("audio_tokens", 0)
        self._usage["completion_tokens_details"][
            "accepted_prediction_tokens"
        ] += completion_tokens_details.get("accepted_prediction_tokens", 0)
        self._usage["completion_tokens_details"][
            "rejected_prediction_tokens"
        ] += completion_tokens_details.get("rejected_prediction_tokens", 0)


def _validate_config_param(name, value):
    # Accept basic JSON types
    if isinstance(value, (str, int, float, bool, type(None))):
        return
    # Recursively check mappings (dict, Mapping)
    if isinstance(value, Mapping):
        for k, v in value.items():
            if not isinstance(k, str):
                raise TypeError(
                    f"Config parameter '{name}' "
                    f"has a non-string key '{k}' of type {type(k).__name__}."
                )
            _validate_config_param(f"{name}.{k}", v)
        return
    # Recursively check sequences (list, tuple, set) but not str/bytes
    if isinstance(value, (list, tuple, set)):
        for idx, item in enumerate(value):
            _validate_config_param(f"{name}[{idx}]", item)
        return
    # Explicitly reject bytes and bytearray
    if isinstance(value, (bytes, bytearray)):
        raise TypeError(
            f"Config parameter '{name}' "
            f"is of type {type(value).__name__}, which is not JSON serializable."
        )
    # Fallback: try json serialization
    try:
        json.dumps(value)
    except Exception:
        raise TypeError(
            f"Config parameter '{name}' "
            f"with value '{value}' of type {type(value).__name__} is not serializable."
        )



# afnio/__init__.py

from afnio import autodiff, cognitive, optim, utils
from afnio._model_client import (
    get_backward_model_client,
    set_backward_model_client,
)
from afnio._variable import Variable
from afnio.autodiff import is_grad_enabled, no_grad, set_grad_enabled
from afnio.autodiff.graph import (
    GradientEdge,
    Node,
)
from afnio.cognitive import functional
from afnio.serialization import load, save

from ._utils import (
    _validate_multi_turn_messages,
    _validate_typed_sequence,
)

__all__ = [
    "GradientEdge",
    "Node",
    "Variable",
    "_validate_multi_turn_messages",
    "_validate_typed_sequence",
    "autodiff",
    "cognitive",
    "functional",
    "get_backward_model_client",
    "is_grad_enabled",
    "load",
    "no_grad",
    "optim",
    "save",
    "set_backward_model_client",
    "set_grad_enabled",
    "utils",
]

# Please keep this list sorted
assert __all__ == sorted(__all__)



from typing import List

from afnio._variable import Variable
from afnio.autodiff.basic_ops import Sum as SumOp

from .module import Module


class Sum(Module):
    """
    Aggregates a list of input Variables into a single output Variable.

    This module utilizes the `Sum` operation from `afnio.autodiff.basic_ops`.
    It supports both numerical (int, float) and string data types. For numerical data,
    it computes the sum. For string data, it concatenates the values and wraps
    each in `<ITEM></ITEM>` tags.

    Note:
        This module does not have any trainable parameters.

    Example:
        >>> import afnio as hf
        >>> from afnio import cognitive as cog
        >>> class Summation(cog.Module):
        ...     def __init__(self):
        ...         super().__init__()
        ...         self.sum = cog.Sum()
        >>>     def forward(self, x):
        ...         return self.sum(x)
        >>> input1 = hf.Variable(data="abc", role="input1")
        >>> input2 = hf.Variable(data="def", role="input2")
        >>> input3 = hf.Variable(data="ghi", role="input3")
        >>> summation = Summation()
        >>> result = summation([input1, input2, input3])
        >>> print(result.data)
        '<ITEM>abc</ITEM><ITEM>def</ITEM><ITEM>ghi</ITEM>'
        >>> print(result.role)
        'input1 and input2 and input3'

    Raises:
        TypeError: If any input is not an instance of `Variable`.

    See Also:
        :class:`afnio.autodiff.basic_ops.Sum` for the underlying operation.
    """

    def __init__(self):
        super().__init__()

    def forward(self, x: List[Variable]) -> Variable:
        if not isinstance(x, list) and all(isinstance(y, Variable) for y in x):
            raise TypeError("All inputs must be instances of 'Variable'.")
        return SumOp.apply(x)



# afnio/cognitive/__init__.py

from .functional import (
    add,
    chat_completion,
    deterministic_evaluator,
    exact_match_evaluator,
    lm_judge_evaluator,
    split,
    sum,
)
from .modules.add import Add
from .modules.chat_completion import ChatCompletion
from .modules.deterministic_evaluator import DeterministicEvaluator
from .modules.exact_match_evaluator import ExactMatchEvaluator
from .modules.lm_judge_evaluator import LMJudgeEvaluator
from .modules.module import Module
from .modules.split import Split
from .modules.sum import Sum
from .parameter import Parameter

__all__ = [
    "Add",
    "ChatCompletion",
    "DeterministicEvaluator",
    "ExactMatchEvaluator",
    "LMJudgeEvaluator",
    "Module",
    "Parameter",
    "Split",
    "Sum",
    "add",
    "chat_completion",
    "deterministic_evaluator",
    "exact_match_evaluator",
    "lm_judge_evaluator",
    "split",
    "sum",
]

# Please keep this list sorted
assert __all__ == sorted(__all__)



import os

import pytest

from afnio.tellurio import _close_singleton_ws_client, login
from afnio.tellurio.client import InvalidAPIKeyError, get_default_client


@pytest.mark.asyncio
async def test_login_success():
    """
    Test the login function with real HTTP and WebSocket connections.
    """
    api_key = os.getenv("TEST_ACCOUNT_API_KEY", "valid_api_key")

    # Call the login function
    result = login(api_key=api_key)

    # Assert the result contains the expected keys
    assert "email" in result
    assert "username" in result
    assert "session_id" in result


@pytest.mark.asyncio
async def test_login_invalid_api_key():
    """
    Test the login function with an invalid API key.
    This should raise a ValueError.
    """
    # Use an invalid API key for testing
    api_key = "invalid_api_key"

    # Call the login function and assert it raises a ValueError
    with pytest.raises(
        InvalidAPIKeyError, match="Login failed due to invalid API key."
    ):
        login(api_key=api_key)


def test_close_singleton_ws_client_direct():
    """
    Test that the singleton WebSocket client is closed directly.
    """
    _, ws_client = get_default_client()
    assert ws_client is not None

    api_key = os.getenv("TEST_ACCOUNT_API_KEY", "valid_api_key")
    login(api_key=api_key)
    assert ws_client.connection is not None

    # Close the singleton WebSocket client
    _close_singleton_ws_client()
    assert ws_client.connection is None



# Define the global default active Run instances

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from afnio.tellurio.run import Run

_active_run = None


def set_active_run(run: "Run"):
    """
    Sets the active run globally.

    Args:
        run (Run): The Run instance to set as active.
    """
    global _active_run
    _active_run = run


def get_active_run() -> "Run":
    """
    Gets the active run.
    If no active run is set, it raises an exception.

    Returns:
        Run: The currently active Run instance.
    """
    global _active_run
    if _active_run is None:
        raise ValueError("No active run is set.")
    return _active_run



from typing import Any, Dict, Optional, Union

from afnio._utils import MultiTurnMessages
from afnio._variable import Variable
from afnio.autodiff.lm_ops import ChatCompletion as ChatCompletionOp
from afnio.models import ChatCompletionModel

from .module import Module


class ChatCompletion(Module):
    """
    Generates a chat-based completion using a language model.

    This module leverages the `ChatCompletion` operation from
    `afnio.autodiff.lm_ops` to perform model inference. The `forward` method
    accepts a list of `messages` representing the conversation history, with optional
    dynamic `inputs` for filling placeholders within the messages. The
    `forward_model_client` is responsible for interfacing with the language model
    (e.g., GPT), while `completion_args` allows customization of generation parameters
    such as temperature, maximum tokens, and seed.

    Example:
        >>> import afnio as hf
        >>> from afnio import cognitive as cog
        >>> from afnio.models.openai import OpenAI
        >>> from afnio import set_backward_model_client
        >>> fwd_model_client = OpenAI()
        >>> fwd_model_args = {"model": "gpt-4o", "temperature": 0.7}
        >>> set_backward_model_client("openai/gpt-4o")
        >>> class Assistant(cog.Module):
        ...     def __init__(self):
        ...         super().__init__()
        ...         self.chat = cog.ChatCompletion()
        ...     def forward(self, fwd_model, messages, inputs, **completion_args):
        ...         return self.chat(fwd_model, messages, inputs, **completion_args)
        >>> system = Variable(
        ...     "You are a helpful assistant.",
        ...     role="system instruction",
        ...     requires_grad=True
        ... )
        >>> user = Variable("Translate 'Hello' to {language}.", role="user query")
        >>> language = hf.Variable("Italian", role="language")
        >>> messages = [
        ...     {"role": "system", "content": [system]},
        ...     {"role": "user", "content": [user]},
        ... ]
        >>> model = Assistant()
        >>> response = model(
        ...     fwd_model_client,
        ...     messages,
        ...     inputs={"language": language},
        ...     **fwd_model_args
        ... )
        >>> print(response.data)
        'Ciao'
        >>> feedback = Variable("Use only capital letters.", role="feedback")
        >>> response.backward(feedback)
        >>> system.grad[0].data
        'The system instruction should enforce the use of capital letters only.'

    See Also:
        :class:`afnio.autodiff.lm_ops.ChatCompletion` for the underlying operation.
    """

    forward_model_client: Optional[ChatCompletionModel]
    messages: MultiTurnMessages
    completion_args: Dict[str, Any]

    def __init__(self):
        super().__init__()

        self.register_model("forward_model_client", None)
        self.register_chat("messages", None)
        self.register_completion_config("completion_args", None)

    def forward(
        self,
        forward_model_client: Optional[ChatCompletionModel],
        messages: MultiTurnMessages,
        inputs: Optional[Dict[str, Union[str, Variable]]] = None,
        **completion_args,
    ) -> Variable:
        self.forward_model_client = forward_model_client
        self.messages = messages
        self.completion_args = completion_args
        return ChatCompletionOp.apply(
            self.forward_model_client,
            self.messages,
            inputs,
            **self.completion_args,
        )



import json
import os

import pytest
from keyring import set_keyring

from afnio.tellurio.client import (
    InvalidAPIKeyError,
    TellurioClient,
)
from afnio.tellurio.utils import get_config_path
from tests.utils import InMemoryKeyring


class TestTellurioClient:
    @pytest.fixture(autouse=True)
    def setup_and_teardown(self):
        """
        Fixture to use the in-memory keyring backend for tests.
        Ensures tests do not interact with the real keyring.
        """
        self.test_keyring = InMemoryKeyring()
        set_keyring(self.test_keyring)
        yield

    def test_login_success(self):
        """
        Test that a valid API key logs in successfully and is stored in the keyring.
        """
        api_key = os.getenv("TEST_ACCOUNT_API_KEY", "valid_api_key")
        username = os.getenv("TEST_USER_USERNAME", "TestUser")
        service = os.getenv("KEYRING_SERVICE_NAME", "Tellurio")

        client = TellurioClient()
        response = client.login(api_key=api_key)

        # Assert that the response is a success message
        assert response == {"email": "test@tellurio.ai", "username": username}

        # Assert that the API key was stored in the test keyring
        stored_api_key = self.test_keyring.get_password(service, username)
        assert stored_api_key == api_key

    def test_login_invalid_api_key(self):
        """
        Test that an invalid API key returns None and does not log in.
        """
        client = TellurioClient()

        with pytest.raises(
            InvalidAPIKeyError, match="Login failed due to invalid API key."
        ):
            client.login(api_key="invalid_api_key")

        with pytest.raises(
            InvalidAPIKeyError, match="Re-login failed due to invalid API key."
        ):
            client.login(api_key="invalid_api_key", relogin=True)

    def test_login_stored_api_key(self):
        """
        Test that a stored API key in the keyring is used for login.
        """
        api_key = os.getenv("TEST_ACCOUNT_API_KEY", "valid_api_key")
        username = os.getenv("TEST_USER_USERNAME", "TestUser")
        service = os.getenv("KEYRING_SERVICE_NAME", "Tellurio")

        # Simulate a stored API key in the test keyring
        self.test_keyring.set_password(service, username, api_key)

        # Write the username to the dummy config file
        with open(get_config_path(), "w") as f:
            json.dump({"username": username}, f)

        # Use the temporary path for testing
        client = TellurioClient()
        response = client.login()

        # Assert that the response is a success message
        assert response == {"email": "test@tellurio.ai", "username": username}

    def test_login_relogin_with_new_api_key(self):
        """
        Test that re-login with a new API key replaces the old key in the keyring.
        """
        username = os.getenv("TEST_USER_USERNAME", "TestUser")
        service = os.getenv("KEYRING_SERVICE_NAME", "Tellurio")

        old_api_key = "old_api_key"
        new_api_key = os.getenv("TEST_ACCOUNT_API_KEY", "valid_api_key")

        # Simulate a stored old API key in the test keyring
        self.test_keyring.set_password(service, username, old_api_key)

        client = TellurioClient()
        response = client.login(api_key=new_api_key, relogin=True)

        # Assert that the response is a success message
        assert response == {"email": "test@tellurio.ai", "username": username}

        # Assert that the new API key was stored in the test keyring
        stored_api_key = self.test_keyring.get_password(service, username)
        assert stored_api_key == new_api_key

    def test_login_relogin_with_missing_api_key(self):
        """
        Test that re-login without providing a new API key raises a ValueError.
        """
        username = os.getenv("TEST_USER_USERNAME", "TestUser")
        service = os.getenv("KEYRING_SERVICE_NAME", "Tellurio")

        # Simulate a stored old API key in the test keyring
        self.test_keyring.set_password(service, username, "old_api_key")

        client = TellurioClient()

        # Attempt re-login without providing a new API key
        with pytest.raises(ValueError, match="API key is required for re-login."):
            client.login(relogin=True)

    def test_invalid_api_key_not_stored(self):
        """
        Test that an invalid API key is not stored in the keyring.
        """
        username = os.getenv("TEST_USER_USERNAME", "TestUser")
        service = os.getenv("KEYRING_SERVICE_NAME", "Tellurio")

        client = TellurioClient()

        # Attempt to log in with an invalid API key
        with pytest.raises(
            InvalidAPIKeyError, match="Login failed due to invalid API key."
        ):
            client.login(api_key="invalid_api_key")

        # Assert that the invalid API key was not stored in the keyring
        stored_api_key = self.test_keyring.get_password(service, username)
        assert stored_api_key is None



from typing import Generic, TypeVar

T_co = TypeVar("T_co", covariant=True)


class Dataset(Generic[T_co]):
    r"""An abstract class representing a :class:`Dataset`.

    All datasets that represent a map from keys to data samples should subclass
    it. All subclasses should overwrite :meth:`__getitem__`, supporting fetching a
    data sample for a given key and :meth:`__len__`, which is expected to return
    the size of the dataset by the default options of
    :class:`~afnio.utils.data.DataLoader`. Subclasses could also optionally
    implement :meth:`__getitems__`, for speedup batched samples loading. This method
    accepts list of indices of samples of batch and returns list of samples.
    """

    def __getitem__(self, index) -> T_co:
        raise NotImplementedError("Subclasses of Dataset should implement __getitem__.")

    def __len__(self):
        raise NotImplementedError("Subclasses of Dataset should implement __len__.")

    # def __getitems__(self, indices: List) -> List[T_co]:
    # Not implemented to prevent false-positives in fetcher check in
    # torch.utils.data._utils.fetch._MapDatasetFetcher



from typing import Any, Callable, List, Optional, Tuple, Union

from afnio._variable import Variable
from afnio.autodiff.evaluator import ExactMatchEvaluator as ExactMatchEvaluatorOp

from .module import Module


class ExactMatchEvaluator(Module):
    """
    Evaluates predictions using an exact match criterion.

    This module leverages the `ExactMatchEvaluator` operation from
    `afnio.autodiff.evaluator` and is a specialized version of the
    `DeterministicEvaluator` that uses an exact matching function to compare the
    `prediction` and `target`. It returns an evaluation `score` (1 for exact match,
    0 otherwise) and an `explanation` describing the evaluation result.

    Example:
        >>> import afnio as hf
        >>> from afnio import cognitive as cog
        >>> from afnio import set_backward_model_client
        >>> set_backward_model_client("openai/gpt-4o")
        >>> class ExactColor(cog.Module):
        ...     def __init__(self):
        ...         super().__init__()
        ...         self.exact_match = cog.ExactMatchEvaluator()
        ...     def forward(self, prediction, target):
        ...         return self.exact_match(prediction, target)
        >>> prediction = hf.Variable(
        ...     data="green",
        ...     role="color prediction",
        ...     requires_grad=True
        ... )
        >>> target = "red"
        >>> model = ExactColor()
        >>> score, explanation = model(prediction, target)
        >>> print(score.data)
        0
        >>> print(explanation.data)
        'The evaluation function, designed for 'exact match', compared the <DATA> fields of the predicted variable and the target variable, resulting in a score: 0.'
        >>> explanation.backward()
        >>> system.grad[0].data
        'Reassess the criteria that led to the initial prediction of 'green'.'

    Raises:
        TypeError: If inputs are not of the correct types.

    See Also:
        :class:`afnio.autodiff.evaluator.ExactMatchEvaluator` for the underlying
        operation.
    """  # noqa: E501

    reduction_fn: Optional[Callable[[List[Any]], Any]]
    reduction_fn_purpose: Optional[Union[str, Variable]]

    def __init__(self):
        super().__init__()

        self.register_function("reduction_fn", None)
        self.register_buffer("reduction_fn_purpose", None)

    def forward(
        self,
        prediction: Variable,
        target: Union[str, List[str], Variable],
        reduction_fn: Optional[Callable[[List[Any]], Any]] = sum,
        reduction_fn_purpose: Optional[Union[str, Variable]] = "summation",
    ) -> Tuple[Variable, Variable]:
        self.reduction_fn = reduction_fn
        self.reduction_fn_purpose = (
            None
            if reduction_fn_purpose is None
            else (
                reduction_fn_purpose
                if isinstance(reduction_fn_purpose, Variable)
                else Variable(reduction_fn_purpose)
            )
        )
        return ExactMatchEvaluatorOp.apply(
            prediction, target, self.reduction_fn, self.reduction_fn_purpose
        )



import os

import pytest
from click.testing import CliRunner
from keyring import set_keyring

from afnio.tellurio.cli import cli
from tests.utils import InMemoryKeyring


@pytest.fixture(autouse=True)
def setup_keyring():
    """
    Fixture to use the in-memory keyring backend for tests.
    Ensures tests do not interact with the real keyring.
    """
    test_keyring = InMemoryKeyring()
    set_keyring(test_keyring)
    yield test_keyring


@pytest.mark.asyncio
async def test_login_success(setup_keyring):
    """
    Test the CLI login command with a valid API key.
    """
    api_key = os.getenv("TEST_ACCOUNT_API_KEY", "valid_api_key")
    username = os.getenv("TEST_USER_USERNAME", "TestUser")
    service = os.getenv("KEYRING_SERVICE_NAME", "Tellurio")

    runner = CliRunner()
    result = runner.invoke(cli, ["login"], input=f"{api_key}\n")
    assert result.exit_code == 0
    assert f"Currently logged in as {username!r}" in result.output
    assert "API key provided and stored securely in local keyring." in result.output

    # Verify the API key is stored in the in-memory keyring
    stored_api_key = setup_keyring.get_password(service, username)
    assert stored_api_key == api_key


def test_login_invalid_api_key(setup_keyring):
    """
    Test the CLI login command with an invalid API key.
    """
    username = os.getenv("TEST_USER_USERNAME", "TestUser")
    service = os.getenv("KEYRING_SERVICE_NAME", "Tellurio")

    runner = CliRunner()
    result = runner.invoke(cli, ["login"], input="invalid_api_key\n")
    assert result.exit_code == 0
    assert "Login failed: Invalid API key. Please try again." in result.output

    # Verify the invalid API key is not stored in the in-memory keyring
    stored_api_key = setup_keyring.get_password(service, username)
    assert stored_api_key is None


def test_login_relogin(setup_keyring):
    """
    Test the CLI login command with the --relogin option.
    """
    username = os.getenv("TEST_USER_USERNAME", "TestUser")
    service = os.getenv("KEYRING_SERVICE_NAME", "Tellurio")

    # Simulate a stored API key in the in-memory keyring
    setup_keyring.set_password("tellurio", "api_key", "old_api_key")

    api_key = os.getenv("TEST_ACCOUNT_API_KEY", "valid_api_key")

    runner = CliRunner()
    result = runner.invoke(cli, ["login", "--relogin"], input=f"{api_key}\n")
    assert result.exit_code == 0
    assert "API key provided and stored securely in local keyring." in result.output
    assert f"Currently logged in as {username!r}" in result.output

    # Verify the new API key is stored in the in-memory keyring
    stored_api_key = setup_keyring.get_password(service, username)
    assert stored_api_key == api_key



# afnio/models/__init__.py

from .model import (
    ChatCompletionModel,
    EmbeddingModel,
    TextCompletionModel,
)

__all__ = [
    "ChatCompletionModel",
    "EmbeddingModel",
    "TextCompletionModel",
]

# Please keep this list sorted
assert __all__ == sorted(__all__)



from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from afnio._utils import MultiTurnMessages
from afnio._variable import Variable
from afnio.autodiff.basic_ops import Add, Split, Sum
from afnio.autodiff.evaluator import (
    DeterministicEvaluator,
    ExactMatchEvaluator,
    LMJudgeEvaluator,
)
from afnio.autodiff.lm_ops import ChatCompletion
from afnio.models import ChatCompletionModel


def add(x: Variable, y: Variable) -> Variable:
    r"""Implements an addition operation for `Variable` instances within the
    `afnio` framework, supporting automatic differentiation.

    The `Add` function supports both scalar and list `.data` fields:
    - **Scalars**: Adds numerical values (`int`, `float`) or concatenates strings.
    - **Lists**: Performs element-wise addition of corresponding elements from the lists.
      Lists must be of the same length.

    It automatically handles type-based operations:
    - For numerical data (`int`, `float`), it performs arithmetic addition.
    - For strings, it concatenates the values.
    - Mixed types (e.g., string and number) are converted appropriately before performing
      the addition.

    This operation also tracks `Variable` dependencies, enabling automatic gradient
    computation through backpropagation.

    Example with scalar inputs:
        >>> x = Variable(data="abc", role="first input", requires_grad=True)
        >>> y = Variable(data="def", role="second input", requires_grad=False)
        >>> result = F.add(x, y)
        >>> result.data
        'abcdef'
        >>> result.role
        'first input and second input'
        >>> result.requires_grad
        True
        >>> g = Variable(data="MY_FEEDBACK", role="add gradient")
        >>> result.backward(g)
        >>> x.grad.data
        'Here is the combined feedback we got for this specific first input and other variables: MY_FEEDBACK'
        >>> x.grad.role
        'feedback to first input'

    Example with batched inputs:
        >>> x = Variable(data=[1, 2, 3], role="first input", requires_grad=True)
        >>> y = Variable(data=[4, 5, 6], role="second input", requires_grad=False)
        >>> result = F.add(x, y)
        >>> result.data
        [5, 7, 9]
        >>> result.role
        'first input and second input'
        >>> result.requires_grad
        True
    """  # noqa: E501
    return Add.apply(x, y)


def sum(x: List[Variable]) -> Variable:
    r"""
    Implements a summation operation for a list of `Variable` instances within the
    `afnio` framework, supporting automatic differentiation.

    The `Sum` function aggregates the `.data`, `.role`, and `.requires_grad` attributes
    of all input `Variable` instances into a single `Variable`. It supports both
    scalar and list `.data` fields:
    - **Scalars**: Computes the arithmetic sum for numerical data (`int`, `float`)
      or concatenates all string values, wrapping each in `<ITEM></ITEM>` tags.
    - **Lists**: Aggregates the corresponding elements of the lists. For numerical
      data, it sums the corresponding elements. For string data, it concatenates them,
      wrapping each element in `<ITEM></ITEM>` tags.

    During backpropagation, the function distributes the gradient to all input
    `Variable` instances that require gradients.

    Example with scalar inputs:
        >>> x = Variable(data="abc", role="first input", requires_grad=True)
        >>> y = Variable(data="def", role="second input", requires_grad=False)
        >>> result = F.sum([x, y])
        >>> result.data
        '<ITEM>abc</ITEM><ITEM>def</ITEM>'
        >>> result.role
        'first input and second input'
        >>> result.requires_grad
        True
        >>> g = Variable(data="MY_FEEDBACK", role="add gradient")
        >>> result.backward(g)
        >>> x.grad.data
        'Here is the combined feedback we got for this specific first input and other variables: MY_FEEDBACK'
        >>> x.grad.role
        'feedback to first input'

    Example with batched inputs:
        >>> x = Variable(data=[1, 2, 3.5], role="first input", requires_grad=True)
        >>> y = Variable(data=[4, 5, 6], role="second input", requires_grad=False)
        >>> result = F.sum([x, y])
        >>> result.data
        [5, 7, 9.5]
        >>> result.role
        'first input and second input'
        >>> result.requires_grad
        True
    """  # noqa: E501
    return Sum.apply(x)


def split(
    x: Variable,
    sep: Optional[Union[str, Variable]] = None,
    maxsplit: Optional[Union[int, Variable]] = -1,
) -> List[Variable]:
    r"""
    Implements a split operation for `Variable` instances within the
    `afnio` framework, supporting automatic differentiation.

    The `Split` function divides the `.data` of the input `Variable` into multiple parts
    using a specified delimiter `sep`. If `maxsplit` is specified, the split operation
    is limited to a maximum number of splits. It handles both scalar and list `.data`
    fields:
    - **Scalars**: The scalar `.data` (a single string) is split into substrings
      based on the specified `sep` and `maxsplit` parameters.
    - **Lists**: Each element of the list `.data` (strings) is split individually. If
      splits of varying lengths occur, shorter splits are automatically padded with
      empty strings to ensure consistent dimensions.

    During backpropagation, feedback is collected and aggregated across all split parts.
    The combined feedback is propagated back to the original input `Variable`, allowing
    for the proper computation of gradients.

    Example with scalar inputs:
        >>> x = Variable(data="afnio is great!", role="sentence", requires_grad=True)
        >>> result = Split.apply(x, sep=" ", maxsplit=1)
        >>> [var.data for var in result]
        ['afnio', 'is great!']
        >>> result[0].role
        'split part 0 of sentence'
        >>> g_1 = Variable(data="MY_FIRST_FEEDBACK", role="gradient")
        >>> g_2 = Variable(data="MY_SECOND_FEEDBACK", role="gradient")
        >>> result[0].backward(g_1, retain_graph=True)
        >>> result[1].backward(g_2)
        >>> x.grad[0].data
        'Here is the combined feedback we got for this specific sentence and other variables: <ITEM>MY_FIRST_FEEDBACK</ITEM><ITEM></ITEM>'
        >>> x.grad[0].role
        'feedback to sentence'
        >>> x.grad[1].data
        'Here is the combined feedback we got for this specific sentence and other variables: <ITEM></ITEM><ITEM>MY_SECOND_FEEDBACK</ITEM>'
        >>> x.grad[1].role
        'feedback to sentence'

    Example with batched inputs:
        >>> x = Variable(
        ...     data=["afnio is great!", "Deep learning"],
        ...     role="sentences",
        ...     requires_grad=True
        ... )
        >>> result = Split.apply(x, sep=" ", maxsplit=2)
        >>> [var.data for var in result]
        [['afnio', 'Deep'], ['is', 'learning'], ['great!', '']]
        >>> g = Variable(data="MY_FEEDBACK", role="gradient")
        >>> result[1].backward(g)
        >>> x.grad[0].data
        'Here is the combined feedback we got for this specific sentences and other variables: <ITEM></ITEM><ITEM>MY_FEEDBACK</ITEM><ITEM></ITEM>'
        >>> x.grad[0].role
        'feedback to sentences'
    """  # noqa: E501
    return Split.apply(x, sep, maxsplit)


def chat_completion(
    forward_model_client: Optional[ChatCompletionModel],
    messages: MultiTurnMessages,
    inputs: Optional[Dict[str, Union[str, List[str], Variable]]] = None,
    **completion_args,
) -> Variable:
    r"""
    Implements a chat completion operation using the specified lanuage model within
    the `afnio` framework, supporting automatic differentiation.

    ### Features:
    - **Mini-Batching**: Processes multiple input dictionaries simultaneously to improve
        throughput.
    - **Asynchronous Execution**: Both the forward and backward passes are optimized to
        run asynchronous calls for each mini-batch, reducing latency.
    - **Gradient Computation**: Supports automatic differentiation for all `Variables`
        in `messages` and `inputs` arguments, maintaining the order of gradients.

    The `ChatCompletion` function generates a `Variable` responses by passing a
    composite prompt, built from a list of `messages` and optional `inputs`, to the
    `forward_model_client`. Each message is a dictionary with a 'role' (e.g., 'system',
    'user') and a list of `Variable` objects as 'content'. `inputs` is a dictionary
    containing strings, list of strings or `Variable`s providing dynamic values to fill
    placeholders within message templates. If `inputs` contain lists of strings or
    `Variable`s which `.data` field is a list, the response's `.data` field will be a
    list, corresponding to the batched results. Otherwise, the `.data` field will be a
    scalar string. Additional behavior, such as temperature or token limits, can be
    customized through `completion_args`.

    Example with scalar inputs:
        >>> system = Variable(
        ...     "You are a helpful assistant.",
        ...     role="system instruction",
        ...     requires_grad=True
        ... )
        >>> user = Variable("Translate 'Hello' to {language}.", role="user query")
        >>> messages = [
        ...     {"role": "system", "content": [system]},
        ...     {"role": "user", "content": [user]},
        ... ]
        >>> inputs = {"language": Variable("Italian", role="language")}
        >>> response = F.chat_completion(
        ...     model_client,
        ...     messages,
        ...     inputs=inputs,
        ...     temperature=0.7
        ... )
        >>> print(response.data)
        'Ciao'
        'Hola'
        >>> feedback = Variable("Use only capital letters.", role="feedback")
        >>> response.backward(feedback)
        >>> system.grad[0].data
        'The system instruction should enforce the use of capital letters only.'

    Example with batched inputs:
        >>> system = Variable(
        ...     "You are a helpful assistant.",
        ...     role="system instruction",
        ...     requires_grad=True
        ... )
        >>> user = Variable("Translate 'Hello' to {language}.", role="user query")
        >>> messages = [
        ...     {"role": "system", "content": [system]},
        ...     {"role": "user", "content": [user]},
        ... ]
        >>> inputs = {
        ...     "language": [
        ...         Variable("Italian", role="language"),
        ...         Variable("Spanish", role="language")
        ...     ]
        ... }
        >>> response = F.chat_completion(
        ...     model_client,
        ...     messages,
        ...     inputs=inputs,
        ...     temperature=0.7
        ... )
        >>> print(response.data)
        ['Ciao', 'Hola']
    """
    return ChatCompletion.apply(
        forward_model_client,
        messages,
        inputs,
        **completion_args,
    )


def lm_judge_evaluator(
    forward_model_client: Optional[ChatCompletionModel],
    messages: MultiTurnMessages,
    prediction: Variable,
    target: Optional[Union[str, List[str], Variable]] = None,
    inputs: Optional[Dict[str, Union[str, Variable]]] = None,
    success_fn: Optional[Callable[[List[Any]], bool]] = None,
    reduction_fn: Optional[Callable[[List[Any]], Any]] = sum,
    reduction_fn_purpose: Optional[Union[str, Variable]] = "summation",
    eval_mode: Union[bool, Variable] = True,
    **completion_args,
) -> Tuple[Variable, Variable]:
    r"""
    Implements an evaluation of a model prediction using a language model (LM) as the
    judge within the `afnio` framework, supporting automatic differentiation.

    This function returns a `score` and an `explanation`, both as `Variable` objects,
    by comparing a `prediction` against a `target` (when present) using a composite
    prompt. The prompt is constructed from a list of `messages` and optional `inputs`,
    which can dynamically populate placeholders in the message templates. The evaluation
    process leverages the specified `forward_model_client` to perform the
    LM-based assessment.

    The `prediction` is a `Variable`. The `target` can be a string, a list of strings,
    or a `Variable`. Similarly, the `inputs` dictionary can include strings, lists of
    strings, or `Variable`s. Each `Variable` passed as an input argument can have either
    a scalar or a list `.data` field, supporting both individual samples and batch
    processing. For batch processing, the lengths of `prediction`, `target`, and any
    batched `inputs` must match.

    The `success_fn` parameter is a user-defined function that returns `True` when
    all predictions evaluated by the LM as Judge are considered successful, and `False`
    otherwise. If `success_fn` returns `True`, the `backward` pass will skip gradient
    calculations and directly return an empty gradient, optimizing computational time.

    If you are processing a batch of predictions and targets, you can use the
    `reduction_fn` to aggregate individual scores (e.g., using `sum` to compute a total
    score). The `reduction_fn_purpose` parameter is a brief description of the
    aggregation’s purpose (e.g., `"summation"`). If you don’t want any aggregation, set
    both `reduction_fn` and `reduction_fn_purpose` to `None`.

    The function operates in two modes controlled by `eval_mode`:
    - **`eval_mode=True` (default)** – Computes gradients for `prediction` only. Use it
      for direct feedback on predictions.
    - **`eval_mode=False`** – Computes gradients for `messages` and `inputs`. Use it to
      optimize the evaluator or align with human evaluation datasets.

    Additional model parameters, such as temperature, max tokens, or seed values, can
    be passed through `completion_args` to customize the LLM's behavior.

    Example with scalar inputs:
        >>> task = Variable(
        ...     "Evaluate if the translation is accurate.",
        ...     role="evaluation task",
        ...     requires_grad=True
        ... )
        >>> format = Variable(
        ...     "Provide 'score' (true/false) and 'explanation' in JSON.",
        ...     role="output format"
        ... )
        >>> user = Variable(
        ...     "<PREDICTION>{prediction}</PREDICTION><TARGET>{target}</TARGET>",
        ...     role="user query"
        ... )
        >>> prediction = Variable(
        ...     "Hola Mundo",
        ...     role="translated text",
        ...     requires_grad=True
        ... )
        >>> target = Variable("Ciao Mondo", role="expected output")
        >>> messages = [
        ...     {"role": "system", "content": [task, format]},
        ...     {"role": "user", "content": [user]}
        ... ]
        >>> score, explanation = F.lm_judge_evaluator(
        ...     model,
        ...     messages,
        ...     prediction,
        ...     target,
        ...     temperature=0.5,
        ... )
        >>> score.data
        False
        >>> explanation.data
        'The translated text is in Spanish, but the expected is in Italian.'
        >>> explanation.backward()
        >>> prediction.grad[0].data
        'The translated text should be in Italian.'

    Example with batched inputs:
        >>> task = Variable(
        ...     "Evaluate if the translation is accurate.",
        ...     role="evaluation task",
        ...     requires_grad=True
        ... )
        >>> format = Variable(
        ...     "Provide 'score' (true/false) and 'explanation' in JSON.",
        ...     role="output format"
        ... )
        >>> user = Variable(
        ...     "<PREDICTION>{prediction}</PREDICTION><TARGET>{target}</TARGET>",
        ...     role="user query"
        ... )
        >>> prediction = Variable(
        ...     data=["Hola Mundo", "Salve a tutti"],
        ...     role="translated text",
        ...     requires_grad=True,
        ... )
        >>> target = ["Ciao Mondo", "Salve a tutti"]
        >>> score, explanation = F.lm_judge_evaluator(
        ...     model,
        ...     messages,
        ...     prediction,
        ...     target,
        ...     reduction_fn=sum,
        ...     reduction_fn_purpose="summation",
        ... )
        >>> score.data
        1
        >>> explanation.data
        'The evaluation function, designed using an LM as the judge, compared the <DATA> fields of the predicted variable and the target variable across all samples in the batch. These scores were then aggregated using the reduction function 'summation', resulting in a final aggregated score: 1.'
        >>> explanation.backward()
        >>> prediction.grad[0].data
        'The translated text should be in Italian.'
    """  # noqa: E501
    return LMJudgeEvaluator.apply(
        forward_model_client,
        messages,
        prediction,
        target,
        inputs,
        success_fn,
        reduction_fn,
        reduction_fn_purpose,
        eval_mode,
        **completion_args,
    )


def deterministic_evaluator(
    prediction: Variable,
    target: Union[str, List[str], Variable],
    eval_fn: Callable[[Variable, Union[str, Variable]], List[Any]],
    eval_fn_purpose: Union[str, Variable],
    success_fn: Optional[Callable[[List[Any]], bool]],
    reduction_fn: Optional[Callable[[List[Any]], Any]],
    reduction_fn_purpose: Optional[Union[str, Variable]],
) -> Tuple[Variable, Variable]:
    """
    Evaluates predictions deterministically using a user-defined evaluation function
    within the `afnio` framework, supporting automatic differentiation.

    The `DeterministicEvaluator` function computes a `score` and an `explanation` based
    on the `prediction` and `target` inputs using a user-defined evaluation function
    (`eval_fn`). The evaluation function's purpose is described by `eval_fn_purpose`.
    Outputs include a numerical or textual score and a textual explanation, both wrapped
    as  `Variable` objects.

    The `prediction` is a `Variable`. The `target` can be a string, a list of strings,
    or a `Variable`. Each `Variable` passed as an input argument can have either
    a scalar or a list `.data` field, supporting both individual samples and batch
    processing. For batch processing, the lengths of `prediction` and `target` must
    match.

    The `success_fn` parameter is a user-defined function that returns `True` when
    all predictions evaluated by `eval_fn` are considered successful, and `False`
    otherwise. If `success_fn` returns `True`, the `backward` pass will skip gradient
    calculations and directly return an empty gradient, optimizing computational time.

    The `reduction_fn` parameter specifies the aggregation function to use for scores
    across a batch of predictions and targets. When specified, the reduction function's
    purpose is described using `reduction_fn_purpose`. If aggregation is not desired,
    set `reduction_fn` and `reduction_fn_purpose` to `None`.

    Example with scalar inputs:
        >>> prediction = Variable(
        ...     data="green",
        ...     role="color prediction",
        ...     requires_grad=True
        ... )
        >>> target = "red"
        >>> def exact_match_fn(p: str, t: str) -> int:
        ...     return 1 if p == t else 0
        >>> score, explanation = F.deterministic_evaluator(
        ...     prediction,
        ...     target,
        ...     exact_match_fn,
        ...     "exact match",
        ... )
        >>> score.data
        0
        >>> explanation.data
        'The evaluation function, designed for 'exact match', compared the <DATA> field of the predicted variable ('green') with the <DATA> field of the target variable ('red'), resulting in a score: 0.'
        >>> explanation.backward()
        >>> prediction.grad[0].data
        'Reassess the criteria that led to the initial prediction of 'green'.'

    Example with batched inputs:
        >>> prediction = Variable(
        ...     data=["green", "blue"],
        ...     role="color prediction",
        ...     requires_grad=True
        ... )
        >>> target = ["red", "blue"]
        >>> def exact_match_fn(p: str, t: str) -> int:
        ...     return 1 if p == t else 0
        >>> score, explanation = F.deterministic_evaluator(
        ...     prediction,
        ...     target,
        ...     exact_match_fn,
        ...     "exact match",
        ...     reduction_fn=sum,
        ...     reduction_fn_purpose="summation"
        ... )
        >>> score.data
        1
        >>> explanation.data
        'The evaluation function, designed for 'exact match', compared the <DATA> fields of the predicted variable and the target variable across all samples in the batch, generating individual scores for each pair. These scores were then aggregated using the reduction function 'summation', resulting in a final aggregated score: 1.'
        >>> explanation.backward()
        >>> prediction.grad[0].data
        'Reassess the criteria that led to the initial prediction of 'green'.'
    """  # noqa: E501
    return DeterministicEvaluator.apply(
        prediction,
        target,
        eval_fn,
        eval_fn_purpose,
        success_fn,
        reduction_fn,
        reduction_fn_purpose,
    )


def exact_match_evaluator(
    prediction: Variable,
    target: Union[str, List[str], Variable],
    reduction_fn: Optional[Callable[[List[Any]], Any]] = sum,
    reduction_fn_purpose: Optional[Union[str, Variable]] = "summation",
) -> Tuple[Variable, Variable]:
    """
    Evaluates predictions using exact matching within the `afnio` framework,
    supporting automatic differentiation.

    The `ExactMatchEvaluator` function computes a `score` and an `explanation` by
    comparing the `data` fields of a `prediction` and a `target` for an exact match.
    For each sample:
    - A score of `1` is assigned for an exact match.
    - A score of `0` is assigned otherwise.

    The `prediction` is a `Variable`. The `target` can be a string, a list of strings,
    or a `Variable`. Each `Variable` passed as an input argument can have either
    a scalar or a list `.data` field, supporting both individual samples and batch
    processing. For batch processing, the lengths of `prediction` and `target` must
    match.

    If batched inputs are provided, the scores can be aggregated using an optional
    `reduction_fn`, such as `sum`. The purpose of the reduction is described using
    `reduction_fn_purpose`. If aggregation is not desired, set `reduction_fn` and
    `reduction_fn_purpose` to `None`.

    Example with scalar inputs:
        >>> prediction = Variable(
        ...     data="green",,
        ...     role="color prediction",
        ...     requires_grad=True
        ... )
        >>> target = "red",
        >>> score, explanation = F.exact_match_evaluator(prediction, target)
        >>> score.data
        0
        >>> explanation.data
        'The evaluation function, designed for 'exact match', compared the <DATA> field of the predicted variable ('green') with the <DATA> field of the target variable ('red'), resulting in a score: 0.'
        >>> explanation.backward()
        >>> prediction.grad[0].data
        'Reassess the criteria that led to the initial prediction of 'green'.'

    Example with batched inputs:
        >>> prediction = Variable(
        ...     data=["green", "blue"],
        ...     role="color prediction",
        ...     requires_grad=True
        ... )
        >>> target = ["red", "blue"]
        >>> score, explanation = F.exact_match_evaluator(prediction, target)
        >>> score.data
        1
        >>> explanation.data
        'The evaluation function, designed for 'exact match', compared the <DATA> fields of the predicted variable and the target variable across all samples in the batch, generating individual scores for each pair. These scores were then aggregated using the reduction function 'summation', resulting in a final aggregated score: 1.'
        >>> explanation.backward()
        >>> prediction.grad[0].data
        'Reassess the criteria that led to the initial prediction of 'green'.'
    """  # noqa: E501
    return ExactMatchEvaluator.apply(
        prediction, target, reduction_fn, reduction_fn_purpose
    )



from typing import Any, Callable, List, Optional, Tuple, Union

from afnio._variable import Variable
from afnio.autodiff.evaluator import (
    DeterministicEvaluator as DeterministicEvaluatorOp,
)

from .module import Module


class DeterministicEvaluator(Module):
    """
    Evaluates predictions deterministically using a user-defined evaluation function.

    This module utilizes the `DeterministicEvaluator` operation from
    `afnio.autodiff.evaluator` to compute evaluation scores and explanations.
    The `forward` method takes in a `prediction`, a `target`, an evaluation
    function (`eval_fn`), and its purpose description (`eval_fn_purpose`). It also
    accepts a reduction function (`reduction_fn`) and its purpose description
    (`reduction_fn_purpose`) to aggregate scores if needed. The method outputs
    an evaluation `score` and an `explanation`, both as `Variable` instances. The
    `success_fn` checks if all evaluations are successful, allowing the `backward` pass
    to skip unnecessary gradient computations. The method outputs an evaluation
    `score` and an `explanation`, both as `Variable` instances.

    Example:
        >>> import afnio as hf
        >>> from afnio import cognitive as cog
        >>> from afnio import set_backward_model_client
        >>> set_backward_model_client("openai/gpt-4o")
        >>> class ExactColor(cog.Module):
        ...     def __init__(self):
        ...         super().__init__()
        ...         def exact_match_fn(pred: str, tgt: str) -> int:
        ...             return 1 if pred == tgt_data else 0
        ...         self.exact_match_fn = exact_match_fn
        ...         self.fn_purpose = "exact match"
        ...         self.reduction_fn = sum
        ...         self.reduction_fn_purpose = "summation"
        ...         self.exact_match = cog.DeterministicEvaluator()
        ...     def forward(self, prediction, target):
        ...         return self.exact_match(
        ...             prediction,
        ...             target,
        ...             self.exact_match_fn,
        ...             self.fn_purpose,
        ...             self.reduction_fn,
        ...             self.reduction_fn_purpose,
        ...         )
        >>> prediction = hf.Variable(
        ...     data=["the color is green", "blue"],
        ...     role="color prediction",
        ...     requires_grad=True
        ... )
        >>> target = ["green", "blue"]
        >>> model = ExactColor()
        >>> score, explanation = model(prediction, target)
        >>> print(score.data)
        1
        >>> print(explanation.data)
        'The evaluation function, designed for 'exact match', compared the <DATA> fields of the predicted variable and the target variable across all samples in the batch, generating individual scores for each pair. These scores were then aggregated using the reduction function 'summation', resulting in a final aggregated score: 1.'
        >>> explanation.backward()
        >>> prediction.grad[0].data
        'Reassess the criteria that led to the initial prediction of 'green'.'

    Raises:
        TypeError: If inputs are not of the correct types.

    See Also:
        :class:`afnio.autodiff.evaluator.DeterministicEvaluator` for the underlying
        operation.
    """  # noqa: E501

    eval_fn: Callable[[Variable, Union[str, Variable]], List[Any]]
    eval_fn_purpose: Union[str, Variable]
    success_fn: Optional[Callable[[List[Any]], bool]]
    reduction_fn: Optional[Callable[[List[Any]], Any]]
    reduction_fn_purpose: Optional[Union[str, Variable]]

    def __init__(self):
        super().__init__()

        self.register_function("eval_fn", None)
        self.register_buffer("eval_fn_purpose", None)
        self.register_function("success_fn", None)
        self.register_function("reduction_fn", None)
        self.register_buffer("reduction_fn_purpose", None)

    def forward(
        self,
        prediction: Variable,
        target: Union[str, List[str], Variable],
        eval_fn: Callable[[Variable, Union[str, Variable]], List[Any]],
        eval_fn_purpose: Union[str, Variable],
        success_fn: Optional[Callable[[List[Any]], bool]],
        reduction_fn: Optional[Callable[[List[Any]], Any]],
        reduction_fn_purpose: Optional[Union[str, Variable]],
    ) -> Tuple[Variable, Variable]:
        self.eval_fn = eval_fn
        self.eval_fn_purpose = (
            None
            if eval_fn_purpose is None
            else (
                eval_fn_purpose
                if isinstance(eval_fn_purpose, Variable)
                else Variable(eval_fn_purpose)
            )
        )
        self.success_fn = success_fn
        self.reduction_fn = reduction_fn
        self.reduction_fn_purpose = (
            None
            if reduction_fn_purpose is None
            else (
                reduction_fn_purpose
                if isinstance(reduction_fn_purpose, Variable)
                else Variable(reduction_fn_purpose)
            )
        )
        return DeterministicEvaluatorOp.apply(
            prediction,
            target,
            self.eval_fn,
            self.eval_fn_purpose,
            self.success_fn,
            self.reduction_fn,
            self.reduction_fn_purpose,
        )
